### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 3
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ##### END GPL LICENSE BLOCK #####

import bpy
import bmesh

from mathutils import Vector
from mathutils.geometry import intersect_point_line

from .drawing_utilities import SnapDrawn
from .common_utilities import snap_utilities
from .common_classes import (
    CharMap,
    Constrain,
    SnapNavigation,
    SnapUtilities,
    )


if not __package__:
    __package__ = "mesh_snap_utilities_line"


def get_closest_edge(bm, point, dist):
    r_edge = None
    for edge in bm.edges:
        v1 = edge.verts[0].co
        v2 = edge.verts[1].co
        # Test the BVH (AABB) first
        for i in range(3):
            if v1[i] <= v2[i]:
                isect = v1[i] - dist <= point[i] <= v2[i] + dist
            else:
                isect = v2[i] - dist <= point[i] <= v1[i] + dist

            if not isect:
                break
        else:
            ret = intersect_point_line(point, v1, v2)

            if ret[1] < 0.0:
                tmp = v1
            elif ret[1] > 1.0:
                tmp = v2
            else:
                tmp = ret[0]

            new_dist = (point - tmp).length
            if new_dist <= dist:
                dist = new_dist
                r_edge = edge

    return r_edge


def get_loose_linked_edges(bmvert):
    linked = [e for e in bmvert.link_edges if not e.link_faces]
    for e in linked:
        linked += [le for v in e.verts if not v.link_faces for le in v.link_edges if le not in linked]
    return linked


def make_line(self, bm_geom, location):
    obj = self.main_snap_obj.data[0]
    bm = self.main_bm
    split_faces = set()

    update_edit_mesh = False

    if bm_geom is None:
        vert = bm.verts.new(location)
        self.list_verts.append(vert)
        update_edit_mesh = True

    elif isinstance(bm_geom, bmesh.types.BMVert):
        if (bm_geom.co - location).length_squared < .001:
            if self.list_verts == [] or self.list_verts[-1] != bm_geom:
                self.list_verts.append(bm_geom)
        else:
            vert = bm.verts.new(location)
            self.list_verts.append(vert)
            update_edit_mesh = True

    elif isinstance(bm_geom, bmesh.types.BMEdge):
        self.list_edges.append(bm_geom)
        ret = intersect_point_line(location, bm_geom.verts[0].co, bm_geom.verts[1].co)

        if (ret[0] - location).length_squared < .001:
            if ret[1] == 0.0:
                vert = bm_geom.verts[0]
            elif ret[1] == 1.0:
                vert = bm_geom.verts[1]
            else:
                edge, vert = bmesh.utils.edge_split(bm_geom, bm_geom.verts[0], ret[1])
                update_edit_mesh = True

            if self.list_verts == [] or self.list_verts[-1] != vert:
                self.list_verts.append(vert)
                self.geom = vert # hack to highlight in the drawing
            # self.list_edges.append(edge)

        else:  # constrain point is near
            vert = bm.verts.new(location)
            self.list_verts.append(vert)
            update_edit_mesh = True

    elif isinstance(bm_geom, bmesh.types.BMFace):
        split_faces.add(bm_geom)
        vert = bm.verts.new(location)
        self.list_verts.append(vert)
        update_edit_mesh = True

    # draw, split and create face
    if len(self.list_verts) >= 2:
        v1, v2 = self.list_verts[-2:]
        edge = bm.edges.get([v1, v2])
        if edge:
                self.list_edges.append(edge)
        else:
            if not v2.link_edges:
                edge = bm.edges.new([v1, v2])
                self.list_edges.append(edge)
            else:  # split face
                v1_link_faces = v1.link_faces
                v2_link_faces = v2.link_faces
                if v1_link_faces and v2_link_faces:
                    split_faces.update(set(v1_link_faces).intersection(v2_link_faces))

                else:
                    if v1_link_faces:
                        faces = v1_link_faces
                        co2 = v2.co.copy()
                    else:
                        faces = v2_link_faces
                        co2 = v1.co.copy()

                    for face in faces:
                        if bmesh.geometry.intersect_face_point(face, co2):
                            co = co2 - face.calc_center_median()
                            if co.dot(face.normal) < 0.001:
                                split_faces.add(face)

                if split_faces:
                    edge = bm.edges.new([v1, v2])
                    self.list_edges.append(edge)
                    ed_list = get_loose_linked_edges(v2)
                    for face in split_faces:
                        facesp = bmesh.utils.face_split_edgenet(face, ed_list)
                    del split_faces
                else:
                    if self.intersect:
                        facesp = bmesh.ops.connect_vert_pair(bm, verts=[v1, v2], verts_exclude=bm.verts)
                        # print(facesp)
                    if not self.intersect or not facesp['edges']:
                        edge = bm.edges.new([v1, v2])
                        self.list_edges.append(edge)
                    else:
                        for edge in facesp['edges']:
                            self.list_edges.append(edge)
            update_edit_mesh = True

        # create face
        if self.create_face:
            ed_list = set(self.list_edges)
            for edge in v2.link_edges:
                if edge not in ed_list and edge.other_vert(v2) in self.list_verts:
                    ed_list.add(edge)
                    break

            ed_list.update(get_loose_linked_edges(v2))

            bmesh.ops.edgenet_fill(bm, edges=list(ed_list))
            update_edit_mesh = True
            # print('face created')

    if update_edit_mesh:
        obj.data.update_gpu_tag()
        obj.data.update_tag()
        obj.update_from_editmode()
        obj.update_tag()
        bmesh.update_edit_mesh(obj.data)
        self.sctx.tag_update_drawn_snap_object(self.main_snap_obj)
        #bm.verts.index_update()

        bpy.ops.ed.undo_push(message="Undo draw line*")

    return [obj.matrix_world @ v.co for v in self.list_verts]


class SnapUtilitiesLine(SnapUtilities, bpy.types.Operator):
    """Make Lines. Connect them to split faces"""
    bl_idname = "mesh.snap_utilities_line"
    bl_label = "Line Tool"
    bl_options = {'REGISTER'}

    wait_for_input : bpy.props.BoolProperty(name="Wait for Input", default=True)

    def _exit(self, context):
        #avoids unpredictable crashes
        del self.main_snap_obj
        del self.main_bm
        del self.list_edges
        del self.list_verts
        del self.list_verts_co

        bpy.types.SpaceView3D.draw_handler_remove(self._handle, 'WINDOW')
        context.area.header_text_set(None)
        self.snap_context_free()

        #Restore initial state
        context.tool_settings.mesh_select_mode = self.select_mode
        context.space_data.overlay.show_face_center = self.show_face_center

    def _init_snap_line_context(self, context):
        self.prevloc = Vector()
        self.list_verts = []
        self.list_edges = []
        self.list_verts_co = []
        self.bool_update = True
        self.vector_constrain = ()
        self.len = 0

        if not (self.bm and self.obj):
            self.obj = context.edit_object
            self.bm = bmesh.from_edit_mesh(self.obj.data)

        self.main_snap_obj = self.snap_obj = self.sctx._get_snap_obj_by_obj(self.obj)
        self.main_bm = self.bm

    def _shift_contrain_callback(self):
        if isinstance(self.geom, bmesh.types.BMEdge):
            mat = self.main_snap_obj.mat
            verts_co = [mat @ v.co for v in self.geom.verts]
            return verts_co[1] - verts_co[0]

    def modal(self, context, event):
        if self.navigation_ops.run(context, event, self.prevloc if self.vector_constrain else self.location):
            return {'RUNNING_MODAL'}

        context.area.tag_redraw()

        if event.ctrl and event.type == 'Z' and event.value == 'PRESS':
            bpy.ops.ed.undo()
            if not self.wait_for_input:
                self._exit(context)
                return {'FINISHED'}
            else:
                del self.bm
                del self.main_bm
                self.charmap.clear()

                bpy.ops.object.mode_set(mode='EDIT') # just to be sure
                bpy.ops.mesh.select_all(action='DESELECT')
                context.tool_settings.mesh_select_mode = (True, False, True)
                context.space_data.overlay.show_face_center = True

                self.snap_context_update(context)
                self._init_snap_line_context(context)
                self.sctx.update_all()

                return {'RUNNING_MODAL'}

        is_making_lines = bool(self.list_verts_co)

        if (event.type == 'MOUSEMOVE' or self.bool_update) and self.charmap.length_entered_value == 0.0:
            mval = Vector((event.mouse_region_x, event.mouse_region_y))

            if self.rv3d.view_matrix != self.rotMat:
                self.rotMat = self.rv3d.view_matrix.copy()
                self.bool_update = True
                snap_utilities.cache.clear()
            else:
                self.bool_update = False

            self.snap_obj, self.prevloc, self.location, self.type, self.bm, self.geom, self.len = snap_utilities(
                    self.sctx,
                    self.main_snap_obj,
                    mval,
                    constrain=self.vector_constrain,
                    previous_vert=(self.list_verts[-1] if self.list_verts else None),
                    increment=self.incremental)

            self.snap_to_grid()

            if is_making_lines and self.preferences.auto_constrain:
                loc = self.list_verts_co[-1]
                vec, type = self.constrain.update(self.sctx.region, self.sctx.rv3d, mval, loc)
                self.vector_constrain = [loc, loc + vec, type]

        if event.value == 'PRESS':
            if is_making_lines and self.charmap.modal_(context, event):
                self.bool_update = self.charmap.length_entered_value == 0.0

                if not self.bool_update:
                    text_value = self.charmap.length_entered_value
                    vector = (self.location - self.list_verts_co[-1]).normalized()
                    self.location = self.list_verts_co[-1] + (vector * text_value)
                    del vector

            elif self.constrain.modal(event, self._shift_contrain_callback):
                self.bool_update = True
                if self.constrain.last_vec:
                    if self.list_verts_co:
                        loc = self.list_verts_co[-1]
                    else:
                        loc = self.location

                    self.vector_constrain = (loc, loc + self.constrain.last_vec, self.constrain.last_type)
                else:
                    self.vector_constrain = None

            elif event.type in {'LEFTMOUSE', 'RET', 'NUMPAD_ENTER'}:
                if event.type == 'LEFTMOUSE' or self.charmap.length_entered_value != 0.0:
                    if not is_making_lines and self.bm:
                        self.main_snap_obj = self.snap_obj
                        self.main_bm = self.bm

                    mat_inv = self.main_snap_obj.mat.inverted_safe()
                    point = mat_inv @ self.location
                    geom2 = self.geom
                    if geom2:
                        geom2.select = False

                    if self.vector_constrain:
                        geom2 = get_closest_edge(self.main_bm, point, .001)

                    self.list_verts_co = make_line(self, geom2, point)

                    self.vector_constrain = None
                    self.charmap.clear()
                else:
                    self._exit(context)
                    return {'FINISHED'}

            elif event.type == 'F8':
                self.vector_constrain = None
                self.constrain.toogle()

            elif event.type in {'RIGHTMOUSE', 'ESC'}:
                if not self.wait_for_input or not is_making_lines or event.type == 'ESC':
                    if self.geom:
                        self.geom.select = True
                    self._exit(context)
                    return {'FINISHED'}
                else:
                    snap_utilities.cache.clear()
                    self.vector_constrain = None
                    self.list_edges = []
                    self.list_verts = []
                    self.list_verts_co = []
                    self.charmap.clear()

        a = ""
        if is_making_lines:
            a = 'length: ' + self.charmap.get_converted_length_str(self.len)

        context.area.header_text_set(text = "hit: %.3f %.3f %.3f %s" % (*self.location, a))

        if True or is_making_lines:
            return {'RUNNING_MODAL'}

        return {'PASS_THROUGH'}

    def draw_callback_px(self):
        if self.bm:
            self.draw_cache.draw_elem(self.snap_obj, self.bm, self.geom)
        self.draw_cache.draw(self.type, self.location, self.list_verts_co, self.vector_constrain, self.prevloc)

    def invoke(self, context, event):
        if context.space_data.type == 'VIEW_3D':
            self.snap_context_init(context)
            self.snap_context_update(context)

            self.constrain = Constrain(self.preferences, context.scene, self.obj)

            self.intersect = self.preferences.intersect
            self.create_face = self.preferences.create_face
            self.navigation_ops = SnapNavigation(context, True)
            self.charmap = CharMap(context)

            self._init_snap_line_context(context)

            # print('name', __name__, __package__)

            #Store current state
            self.select_mode = context.tool_settings.mesh_select_mode[:]
            self.show_face_center = context.space_data.overlay.show_face_center

            #Modify the current state
            bpy.ops.mesh.select_all(action='DESELECT')
            context.tool_settings.mesh_select_mode = (True, False, True)
            context.space_data.overlay.show_face_center = True

            #Store values from 3d view context
            self.rv3d = context.region_data
            self.rotMat = self.rv3d.view_matrix.copy()
            # self.obj_matrix.transposed())

            #modals
            context.window_manager.modal_handler_add(self)

            if not self.wait_for_input:
                mat_inv = self.obj.matrix_world.inverted_safe()
                point = mat_inv @ self.location
                self.list_verts_co = make_line(self, self.geom, point)

            self._handle = bpy.types.SpaceView3D.draw_handler_add(self.draw_callback_px, (), 'WINDOW', 'POST_VIEW')

            return {'RUNNING_MODAL'}
        else:
            self.report({'WARNING'}, "Active space must be a View3d")
            return {'CANCELLED'}


def register():
    bpy.utils.register_class(SnapUtilitiesLine)

if __name__ == "__main__":
    register()



# gpl author: Ryan Inch (Imaginer)

import bpy

get_addon_name = 'space_view3d_brush_menus'

# Property Icon Width
PIW = '       '


# check for (currently) brushes being linked
def get_brush_link(context, types="brush"):
    tool_settings = context.tool_settings
    has_brush = None

    if get_mode() == 'SCULPT':
        datapath = tool_settings.sculpt

    elif get_mode() == 'VERTEX_PAINT':
        datapath = tool_settings.vertex_paint

    elif get_mode() == 'WEIGHT_PAINT':
        datapath = tool_settings.weight_paint

    elif get_mode() == 'TEXTURE_PAINT':
        datapath = tool_settings.image_paint
    else:
        datapath = None

    if types == "brush":
        has_brush = getattr(datapath, "brush", None)

    return has_brush


# Addon settings
def addon_settings():
    # separate function just for more convenience
    addon = bpy.context.preferences.addons[get_addon_name]
    colum_n = addon.preferences.column_set if addon else 1

    return colum_n


def error_handlers(self, op_name, error, reports="ERROR", func=False):
    if self and reports:
        self.report({'WARNING'}, reports + " (See Console for more info)")

    is_func = "Function" if func else "Operator"
    print("\n[Sculpt/Paint Brush Menus]\n{}: {}\nError: {}\n".format(is_func, op_name, error))


# Object modes:
# 'OBJECT' 'EDIT' 'SCULPT'
# 'VERTEX_PAINT' 'WEIGHT_PAINT' 'TEXTURE_PAINT'
# 'PARTICLE_EDIT' 'POSE' 'GPENCIL_EDIT'
def get_mode():
    return bpy.context.object.mode

def menuprop(item, name, value, data_path,
             icon='NONE', disable=False, disable_icon=None,
             custom_disable_exp=None, method=None, path=False):

    # disable the ui
    if disable:
        disabled = False

        # used if you need a custom expression to disable the ui
        if custom_disable_exp:
            if custom_disable_exp[0] == custom_disable_exp[1]:
                item.enabled = False
                disabled = True

        # check if the ui should be disabled for numbers
        elif isinstance(eval("bpy.context.{}".format(data_path)), float):
            if round(eval("bpy.context.{}".format(data_path)), 2) == value:
                item.enabled = False
                disabled = True

        # check if the ui should be disabled for anything else
        else:
            if eval("bpy.context.{}".format(data_path)) == value:
                item.enabled = False
                disabled = True

        # change the icon to the disable_icon if the ui has been disabled
        if disable_icon and disabled:
            icon = disable_icon

    # creates the menu item
    prop = item.operator("wm.context_set_value", text=name, icon=icon)

    # sets what the menu item changes
    if path:
        prop.value = value
        value = eval(value)

    elif type(value) == str:
        prop.value = "'{}'".format(value)

    else:
        prop.value = '{}'.format(value)

    # sets the path to what is changed
    prop.data_path = data_path



# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####
#
# <pep8 compliant>
#
# -----------------------------------------------------------------------
# Author: Alan Odom (Clockmender), Rune Morling (ermo) Copyright (c) 2019
# -----------------------------------------------------------------------
#
# ----------------------------------------------
# Define Addon info
# ----------------------------------------------
#
bl_info = {
    "name": "Precision Drawing Tools (PDT)",
    "author": "Alan Odom (Clockmender), Rune Morling (ermo)",
    "version": (1, 5, 1),
    "blender": (2, 90, 0),
    "location": "View3D > UI > PDT",
    "description": "Precision Drawing Tools for Acccurate Modelling",
    "warning": "",
    "doc_url": "https://github.com/Clockmender/Precision-Drawing-Tools/wiki",
    "category": "3D View",
}


# ----------------------------------------------
# Import modules
# ----------------------------------------------
if "bpy" in locals():
    import importlib

    importlib.reload(pdt_design)
    importlib.reload(pdt_pivot_point)
    importlib.reload(pdt_menus)
    importlib.reload(pdt_library)
    importlib.reload(pdt_view)
    importlib.reload(pdt_xall)
    importlib.reload(pdt_bix)
    importlib.reload(pdt_etof)
    importlib.reload(pdt_tangent)
    importlib.reload(pdt_trig_waves)
else:
    from . import pdt_design
    from . import pdt_pivot_point
    from . import pdt_menus
    from . import pdt_library
    from . import pdt_view
    from . import pdt_xall
    from . import pdt_bix
    from . import pdt_etof
    from . import pdt_tangent
    from . import pdt_trig_waves

import bpy
import os
from pathlib import Path
from bpy.types import (
    AddonPreferences,
    PropertyGroup, Scene,
    WindowManager,
    Object,
)
from bpy.props import (
    BoolProperty,
    CollectionProperty,
    EnumProperty,
    FloatProperty,
    FloatVectorProperty,
    IntProperty,
    PointerProperty,
    StringProperty,
)
from .pdt_msg_strings import (
    PDT_DES_COORDS,
    PDT_DES_FILLETPROF,
    PDT_DES_FILLETRAD,
    PDT_DES_FILLETSEG,
    PDT_DES_FILLETVERTS,
    PDT_DES_FILLINT,
    PDT_DES_FLIPANG,
    PDT_DES_FLIPPER,
    PDT_DES_LIBCOLS,
    PDT_DES_LIBMATS,
    PDT_DES_LIBMODE,
    PDT_DES_LIBOBS,
    PDT_DES_LIBSER,
    PDT_DES_MOVESEL,
    PDT_DES_OBORDER,
    PDT_DES_OFFANG,
    PDT_DES_OFFDIS,
    PDT_DES_OFFPER,
    PDT_DES_OPMODE,
    PDT_DES_OUTPUT,
    PDT_DES_PIVOTDIS,
    PDT_DES_PPLOC,
    PDT_DES_PPSCALEFAC,
    PDT_DES_PPSIZE,
    PDT_DES_PPTRANS,
    PDT_DES_PPWIDTH,
    PDT_DES_ROTMOVAX,
    PDT_DES_TRIM,
    PDT_DES_VALIDLET,
    PDT_DES_WORPLANE,
    PDT_DES_TANCEN1,
    PDT_DES_TANCEN2,
    PDT_DES_TANCEN3,
    PDT_DES_RADIUS1,
    PDT_DES_RADIUS2,
    PDT_DES_TPOINT,
    PDT_DES_EXPCOLL,
    PDT_DES_TANMODE,
)
from .pdt_command import command_run
from .pdt_functions import scale_set


# Declare enum items variables
#
_pdt_obj_items = []
_pdt_col_items = []
_pdt_mat_items = []


class PDTPreferences(AddonPreferences):
    # This must match the addon name, use '__package__'
    # when defining this in a submodule of a python package.

    bl_idname = __name__

    debug: BoolProperty(
        name="Enable console debug output from PDT scripts",
        default=False,
        description="NOTE: Does not enable debugging globally in Blender (only in PDT scripts)",
    )

    pdt_ui_width: IntProperty(
        name="UI Width Cut-off",
        default=350,
        description="Cutoff width for shrinking items per line in menus",
    )

    pdt_input_round: IntProperty(
        name="Input Rounding", default=5, description="Rounding Factor for Inputs"
    )

    def draw(self, context):
        layout = self.layout

        box = layout.box()
        row1 = box.row()
        row2 = box.row()
        row1.prop(self, "debug")
        row2.prop(self, "pdt_ui_width")
        row2.prop(self, "pdt_input_round")


def enumlist_objects(self, context):
    """Populate Objects List from Parts Library.

    Creates list of objects that optionally have search string contained in them
    to populate variable pdt_lib_objects enumerator.

    Args:
        context: Blender bpy.context instance.

    Returns:
        list of Object Names.
    """

    scene = context.scene
    pg = scene.pdt_pg
    file_path = pg.pdt_library_path
    path = Path(bpy.path.abspath(file_path))
    _pdt_obj_items.clear()

    if path.is_file() and ".blend" in str(path):
        with bpy.data.libraries.load(str(path)) as (data_from, _):
            if len(pg.object_search_string) == 0:
                object_names = [obj for obj in data_from.objects]
            else:
                object_names = [obj for obj in data_from.objects if pg.object_search_string in obj]
        for object_name in object_names:
            _pdt_obj_items.append((object_name, object_name, ""))
    else:
        _pdt_obj_items.append(("MISSING", "Library Not Set", ""))
    return _pdt_obj_items


def enumlist_collections(self, context):
    """Populate Collections List from Parts Library.

    Creates list of collections that optionally have search string contained in them
    to populate variable pg.lib_collections enumerator

    Args:
        context: Blender bpy.context instance.

    Returns:
        list of Collections Names.
    """

    scene = context.scene
    pg = scene.pdt_pg
    file_path = pg.pdt_library_path
    path = Path(bpy.path.abspath(file_path))
    _pdt_col_items.clear()

    if path.is_file() and ".blend" in str(path):
        with bpy.data.libraries.load(str(path)) as (data_from, _):
            if len(pg.collection_search_string) == 0:
                object_names = [obj for obj in data_from.collections]
            else:
                object_names = [
                    obj for obj in data_from.collections if pg.collection_search_string in obj
                ]
        for object_name in object_names:
            _pdt_col_items.append((object_name, object_name, ""))
    else:
        _pdt_col_items.append(("MISSING", "Library Not Set", ""))
    return _pdt_col_items


def enumlist_materials(self, context):
    """Populate Materials List from Parts Library.

    Creates list of materials that optionally have search string contained in them
    to populate variable pg.lib_materials enumerator.

    Args:
        context: Blender bpy.context instance.

    Returns:
        list of Object Names.
    """

    scene = context.scene
    pg = scene.pdt_pg
    file_path = pg.pdt_library_path
    path = Path(bpy.path.abspath(file_path))
    _pdt_mat_items.clear()

    if path.is_file() and ".blend" in str(path):
        with bpy.data.libraries.load(str(path)) as (data_from, _):
            if len(pg.material_search_string) == 0:
                object_names = [obj for obj in data_from.materials]
            else:
                object_names = [
                    obj for obj in data_from.materials if pg.material_search_string in obj
                ]
        for object_name in object_names:
            _pdt_mat_items.append((object_name, object_name, ""))
    else:
        _pdt_mat_items.append(("MISSING", "Library Not Set", ""))
    return _pdt_mat_items


class PDTSceneProperties(PropertyGroup):
    """Contains all PDT related properties."""

    pdt_library_path: StringProperty(
        name="Library",
        default="",
        description="Parts Library File",
        maxlen=1024,
        subtype="FILE_PATH",
    )

    object_search_string: StringProperty(name="Search", default="", description=PDT_DES_LIBSER)
    collection_search_string: StringProperty(name="Search", default="", description=PDT_DES_LIBSER)
    material_search_string: StringProperty(name="Search", default="", description=PDT_DES_LIBSER)

    cartesian_coords: FloatVectorProperty(
        name="Coords", default=(0.0, 0.0, 0.0), subtype="XYZ", description=PDT_DES_COORDS
    )
    distance: FloatProperty(
        name="Distance", default=0.0, precision=5, description=PDT_DES_OFFDIS, unit="LENGTH"
    )
    angle: FloatProperty(
        name="Angle", min=-180, max=180, default=0.0, precision=5, description=PDT_DES_OFFANG
    )
    percent: FloatProperty(name="Percent", default=0.0, precision=5, description=PDT_DES_OFFPER)
    plane: EnumProperty(
        items=(
            ("XZ", "Front(X-Z)", "Use X-Z Plane"),
            ("XY", "Top(X-Y)", "Use X-Y Plane"),
            ("YZ", "Right(Y-Z)", "Use Y-Z Plane"),
            ("LO", "View", "Use View Plane"),
        ),
        name="Working Plane",
        default="XZ",
        description=PDT_DES_WORPLANE,
    )
    select: EnumProperty(
        items=(
            ("REL", "Current Pos.", "Move Relative to Current Position"),
            (
                "SEL",
                "Selected Entities",
                "Move Relative to Selected Object or Vertex (Cursor & Pivot Only)",
            ),
        ),
        name="Move Mode",
        default="SEL",
        description=PDT_DES_MOVESEL,
    )
    operation: EnumProperty(
        items=(
            ("CU", "Move Cursor", "This function will Move the Cursor"),
            ("PP", "Move Pivot Point", "This function will Move the Pivot Point"),
            ("MV", "Move Entities", "This function will Move selected Vertices or Objects"),
            ("NV", "Add New Vertex", "This function will Add a New Vertex"),
            (
                "EV",
                "Extrude Vertex/Vertices",
                "This function will Extrude Vertex/Vertices Only in EDIT Mode",
            ),
            ("SE", "Split Edges", "This function will Split Edges Only in EDIT Mode"),
            (
                "DG",
                "Duplicate Geometry",
                "This function will Duplicate Geometry in EDIT Mode (Delta & Direction Only)",
            ),
            (
                "EG",
                "Extrude Geometry",
                "This function will Extrude Geometry in EDIT Mode (Delta & Direction Only)",
            ),
        ),
        name="Operation",
        default="CU",
        description=PDT_DES_OPMODE,
    )
    taper: EnumProperty(
        items=(
            ("RX-MY", "RotX-MovY", "Rotate X - Move Y"),
            ("RX-MZ", "RotX-MovZ", "Rotate X - Move Z"),
            ("RY-MX", "RotY-MovX", "Rotate Y - Move X"),
            ("RY-MZ", "RotY-MovZ", "Rotate Y - Move Z"),
            ("RZ-MX", "RotZ-MovX", "Rotate Z - Move X"),
            ("RZ-MY", "RotZ-MovY", "Rotate Z - Move Y"),
        ),
        name="Axes",
        default="RX-MY",
        description=PDT_DES_ROTMOVAX,
    )

    flip_angle: BoolProperty(name="Flip Angle", default=False, description=PDT_DES_FLIPANG)
    flip_percent: BoolProperty(name="Flip %", default=False, description=PDT_DES_FLIPPER)

    extend: BoolProperty(name="Trim/Extend All", default=False, description=PDT_DES_TRIM)

    lib_objects: EnumProperty(items=enumlist_objects, name="Objects", description=PDT_DES_LIBOBS)
    lib_collections: EnumProperty(
        items=enumlist_collections, name="Collections", description=PDT_DES_LIBCOLS
    )
    lib_materials: EnumProperty(
        items=enumlist_materials, name="Materials", description=PDT_DES_LIBMATS
    )
    lib_mode: EnumProperty(
        items=(
            ("OBJECTS", "Objects", "Use Objects"),
            ("COLLECTIONS", "Collections", "Use Collections"),
            ("MATERIALS", "Materials", "Use Materials"),
        ),
        name="Mode",
        default="OBJECTS",
        description=PDT_DES_LIBMODE,
    )

    rotation_coords: FloatVectorProperty(
        name="Rotation", default=(0.0, 0.0, 0.0), subtype="XYZ", description="Rotation Coordinates"
    )

    object_order: EnumProperty(
        items=(
            ("1,2,3,4", "1,2,3,4", "Objects 1 & 2 are First Line"),
            ("1,3,2,4", "1,3,2,4", "Objects 1 & 3 are First Line"),
            ("1,4,2,3", "1,4,2,3", "Objects 1 & 4 are First Line"),
        ),
        name="Order",
        default="1,2,3,4",
        description=PDT_DES_OBORDER,
    )
    vrotangle: FloatProperty(name="View Rotate Angle", default=10, max=180, min=-180)
    command: StringProperty(
        name="Command", default="CA0,0,0", update=command_run, description=PDT_DES_VALIDLET,
    )
    maths_output: FloatProperty(
        name="Maths output", default=0, description=PDT_DES_OUTPUT,
    )
    error: StringProperty(name="Error", default="")

    # Was pivot* -- is now pivot_*
    pivot_loc: FloatVectorProperty(
        name="Pivot Location", default=(0.0, 0.0, 0.0), subtype="XYZ", description=PDT_DES_PPLOC,
    )
    pivot_scale: FloatVectorProperty(
        name="Pivot Scale", default=(1.0, 1.0, 1.0), subtype="XYZ", description=PDT_DES_PPSCALEFAC
    )
    pivot_size: FloatProperty(
        name="Pivot Factor", min=0.4, max=10, default=2, precision=1, description=PDT_DES_PPSIZE
    )
    pivot_width: IntProperty(name="Width", min=1, max=5, default=2, description=PDT_DES_PPWIDTH)

    pivot_ang: FloatProperty(name="Pivot Angle", min=-180, max=180, default=0.0)

    pivot_dis: FloatProperty(
        name="Pivot Dist", default=0.0, min=0, update=scale_set, description=PDT_DES_PIVOTDIS,
    )
    pivot_alpha: FloatProperty(
        name="Alpha", min=0.2, max=1, default=0.6, precision=1, description=PDT_DES_PPTRANS,
    )
    pivot_show: BoolProperty()

    # Was filletrad
    fillet_radius: FloatProperty(
        name="Fillet Radius", min=0.0, default=1.0, unit="LENGTH", description=PDT_DES_FILLETRAD
    )
    # Was filletnum
    fillet_segments: IntProperty(
        name="Fillet Segments", min=1, default=4, description=PDT_DES_FILLETSEG
    )
    # Was filletpro
    fillet_profile: FloatProperty(
        name="Fillet Profile", min=0.0, max=1.0, default=0.5, description=PDT_DES_FILLETPROF
    )
    # Was filletbool
    fillet_vertices_only: BoolProperty(
        name="Fillet Vertices Only", default=True, description=PDT_DES_FILLETVERTS,
    )
    fillet_intersect: BoolProperty(
        name="Intersect", default=False, description=PDT_DES_FILLINT,
    )
    tangent_point0: FloatVectorProperty(
        name="Coordst1", default=(0.0, 0.0, 0.0), subtype="XYZ", description=PDT_DES_TANCEN1
    )
    tangent_point1: FloatVectorProperty(
        name="Coordst2", default=(0.0, 0.0, 0.0), subtype="XYZ", description=PDT_DES_TANCEN2
    )
    tangent_radius0: FloatProperty(
        name="Arc Radius 1", min=0.00001,  default=1, unit="LENGTH", description=PDT_DES_RADIUS1
    )
    tangent_radius1: FloatProperty(
        name="Arc Radius 2", min=0.00001, default=1, unit="LENGTH", description=PDT_DES_RADIUS2
    )
    tangent_point2: FloatVectorProperty(
        name="Coordst3", default=(0.0, 0.0, 0.0), subtype="XYZ", description=PDT_DES_TANCEN3
    )
    menu_expand: BoolProperty(
        name="Expand", default=False, description=PDT_DES_EXPCOLL,
    )
    tangent_mode: EnumProperty(
        items=(
            ("inner", "Inner", "Inner Tangents"),
            ("outer", "Outer", "Outer Tangents"),
            ("both", "Inner & Outer", "Inner & Outer Tangents"),
            ("point", "From Point", "Tangents from Point"),
        ),
        name="Working Plane",
        default="both",
        description=PDT_DES_TANMODE,
    )

    # For Trig Waves
    trig_type : EnumProperty(
        items=(
            ("sin", "Sine", "Sine Wave"),
            ("cos", "Cosine", "Cosine Wave"),
            ("tan", "Tangent", "Tangent Wave"),
        ),
        name="Wave Form",
        default="sin",
        description="Trig. Wave Form",
    )
    trig_cycles : IntProperty(name="Cycles #", default=1, min=1,
        description="1 Cycle = 180 Degrees")
    trig_amp : FloatProperty(name="Amplitude", default=1, min=0.01,
        description="Maximum Height of 1 Cycle (forms Basis for Tangents)")
    trig_len : FloatProperty(name="Cycle Length", default=2, min=0.02,
        description="Length in Blender Units of 1 Cycle")
    trig_obj : PointerProperty(name="Object", type=Object)
    trig_del : BoolProperty(name="Empty Object", default=False,
        description="Delete ALL Vertices in Object First")
    trig_res : IntProperty(name="Resolution", default=18, min=4, max=72,
        description="Number of Vertices per Cycle (180 Degrees)")
    trig_tanmax : FloatProperty(name="Tangent Max", default=10, min=0.1,
        description="Maximum Permitted Tangent Value")
    trig_off : FloatVectorProperty(name="Start Location", default=(0,0,0),
        description="Location in World Space for Origin of Wave")
    trig_abs : BoolProperty(name="Absolute", default=False,
        description="Use Absolute Values Only")


# List of All Classes in the Add-on to register
#
# Due to the way PropertyGroups work, this needs to be listed/loaded first
# (and unloaded last)
#
classes = (
    PDTPreferences,
    PDTSceneProperties,
    pdt_bix.PDT_OT_LineOnBisection,
    pdt_command.PDT_OT_CommandReRun,
    pdt_design.PDT_OT_PlacementAbs,
    pdt_design.PDT_OT_PlacementDelta,
    pdt_design.PDT_OT_PlacementDis,
    pdt_design.PDT_OT_PlacementView,
    pdt_design.PDT_OT_PlacementCen,
    pdt_design.PDT_OT_PlacementPer,
    pdt_design.PDT_OT_PlacementNormal,
    pdt_design.PDT_OT_PlacementInt,
    pdt_design.PDT_OT_JoinVerts,
    pdt_design.PDT_OT_Angle2,
    pdt_design.PDT_OT_Angle3,
    pdt_design.PDT_OT_Origin,
    pdt_design.PDT_OT_Taper,
    pdt_design.PDT_OT_Fillet,
    pdt_etof.PDT_OT_EdgeToFace,
    pdt_library.PDT_OT_Append,
    pdt_library.PDT_OT_Link,
    pdt_library.PDT_OT_LibShow,
    pdt_menus.PDT_PT_PanelDesign,
    pdt_menus.PDT_PT_PanelTools,
    pdt_menus.PDT_PT_PanelTangent,
    pdt_menus.PDT_PT_PanelCommandLine,
    pdt_menus.PDT_PT_PanelViewControl,
    pdt_menus.PDT_PT_PanelPivotPoint,
    pdt_menus.PDT_PT_PanelPartsLibrary,
    pdt_menus.PDT_PT_PanelTrig,
    pdt_pivot_point.PDT_OT_ModalDrawOperator,
    pdt_pivot_point.PDT_OT_ViewPlaneRotate,
    pdt_pivot_point.PDT_OT_ViewPlaneScale,
    pdt_pivot_point.PDT_OT_PivotToCursor,
    pdt_pivot_point.PDT_OT_CursorToPivot,
    pdt_pivot_point.PDT_OT_PivotSelected,
    pdt_pivot_point.PDT_OT_PivotOrigin,
    pdt_pivot_point.PDT_OT_PivotWrite,
    pdt_pivot_point.PDT_OT_PivotRead,
    pdt_tangent.PDT_OT_TangentOperate,
    pdt_tangent.PDT_OT_TangentOperateSel,
    pdt_tangent.PDT_OT_TangentSet1,
    pdt_tangent.PDT_OT_TangentSet2,
    pdt_tangent.PDT_OT_TangentSet3,
    pdt_tangent.PDT_OT_TangentSet4,
    pdt_tangent.PDT_OT_TangentExpandMenu,
    pdt_trig_waves.PDT_OT_WaveGenerator,
    pdt_view.PDT_OT_ViewRot,
    pdt_view.PDT_OT_ViewRotL,
    pdt_view.PDT_OT_ViewRotR,
    pdt_view.PDT_OT_ViewRotU,
    pdt_view.PDT_OT_ViewRotD,
    pdt_view.PDT_OT_ViewRoll,
    pdt_view.PDT_OT_ViewIso,
    pdt_view.PDT_OT_Reset3DView,
    pdt_xall.PDT_OT_IntersectAllEdges,
)


def register():
    """Register Classes and Create Scene Variables.

    Operates on the classes list defined above.
    """

    from bpy.utils import register_class

    for cls in classes:
        register_class(cls)

    # OpenGL flag
    #
    window_manager = WindowManager
    # Register Internal OpenGL Property
    #
    window_manager.pdt_run_opengl = BoolProperty(default=False)

    Scene.pdt_pg = PointerProperty(type=PDTSceneProperties)


def unregister():
    """Unregister Classes and Delete Scene Variables.

    Operates on the classes list defined above.
    """

    from bpy.utils import unregister_class

    # remove OpenGL data
    pdt_pivot_point.PDT_OT_ModalDrawOperator.handle_remove(
        pdt_pivot_point.PDT_OT_ModalDrawOperator, bpy.context
    )
    window_manager = bpy.context.window_manager
    pdt_wm = "pdt_run_opengl"
    if pdt_wm in window_manager:
        del window_manager[pdt_wm]

    for cls in reversed(classes):
        unregister_class(cls)

    del Scene.pdt_pg


if __name__ == "__main__":
    register()



import bpy

from bpy.types import (
    AddonPreferences,
    PropertyGroup,
    )
from bpy.props import (
    StringProperty,
    BoolProperty,
    EnumProperty,
    IntProperty,
    FloatProperty
    )
from math import radians

from .enum_values import *

# Addon Preferences
class VIEW3D_MT_materialutilities_preferences(AddonPreferences):
    bl_idname = __package__

    new_material_name: StringProperty(
            name = "New Material name",
            description = "What Base name pattern to use for a new created Material\n"
                          "It is appended by an automatic numeric pattern depending\n"
                          "on the number of Scene's materials containing the Base",
            default = "Unnamed Material",
            )
    override_type: EnumProperty(
            name = 'Assignment method',
            description = '',
            items = mu_override_type_enums
            )
    fake_user: EnumProperty(
            name = "Set Fake User",
            description = "Default option for the Set Fake User (Turn fake user on or off)",
            items = mu_fake_user_set_enums,
            default = 'TOGGLE'
            )
    fake_user_affect: EnumProperty(
            name = "Affect",
            description = "Which materials of objects to affect",
            items = mu_fake_user_affect_enums,
            default = 'UNUSED'
            )
    link_to: EnumProperty(
            name = "Change Material Link To",
            description = "Default option for the Change Material Link operator",
            items = mu_link_to_enums,
            default = 'OBJECT'
            )
    link_to_affect: EnumProperty(
            name = "Affect",
            description = "Which materials of objects to affect by default with Change Material Link",
            items = mu_link_affect_enums,
            default = 'SELECTED'
            )
    search_show_limit: IntProperty(
            name = "Show 'Search' Limit",
            description = "How many materials should there be before the 'Search' option is shown "
                          "in the Assign Material and Select By Material menus\n"
                          "Set it to 0 to always show 'Search'",
            min = 0,
            default = 0
            )

    set_smooth_affect: EnumProperty(
            name = "Set Auto Smooth Affect",
            description = "Which objects to affect",
            items = mu_affect_enums,
            default = 'SELECTED'
            )
    auto_smooth_angle: FloatProperty(
            name = "Auto Smooth Angle",
            description = "Maximum angle between face normals that will be considered as smooth",
            subtype = 'ANGLE',
            min = 0,
            max = radians(180),
            default = radians(35)
            )

    def draw(self, context):
        layout = self.layout
        layout.use_property_split = True

        box = layout.box()
        box.label(text = "Defaults")

        a = box.box()
        a.label(text = "Assign Material")
        a.prop(self, "new_material_name", icon = "MATERIAL")
        a.prop(self, "override_type", expand = False)

        b = box.box()
        b.label(text = "Set Fake User")
        b.row().prop(self, "fake_user", expand = False)
        b.row().prop(self, "fake_user_affect", expand = False)

        c = box.box()
        c.label(text = "Set Link To")
        c.row().prop(self, "link_to", expand = False)
        c.row().prop(self, "link_to_affect", expand = False)

        d = box.box()
        d.label(text = "Set Auto Smooth")
        d.row().prop(self, "auto_smooth_angle", expand = False)
        d.row().prop(self, "set_smooth_affect", expand = False)

        box = layout.box()
        box.label(text = "Miscellaneous")

        #col = box.column()
        #row = col.split(factor = 0.5)
        box.prop(self, "search_show_limit", expand = False)


def materialutilities_get_preferences(context):
    return context.preferences.addons[__package__].preferences



import bpy
import bmesh


def main(context):
    obj = context.active_object
    me = obj.data
    bm = bmesh.from_edit_mesh(me)

    uv_layer = bm.loops.layers.uv.verify()

    # adjust uv coordinates
    for face in bm.faces:
        for loop in face.loops:
            loop_uv = loop[uv_layer]
            # use xy position of the vertex as a uv coordinate
            loop_uv.uv = loop.vert.co.xy

    bmesh.update_edit_mesh(me)


class UvOperator(bpy.types.Operator):
    """UV Operator description"""
    bl_idname = "uv.simple_operator"
    bl_label = "Simple UV Operator"

    @classmethod
    def poll(cls, context):
        obj = context.active_object
        return obj and obj.type == 'MESH' and obj.mode == 'EDIT'

    def execute(self, context):
        main(context)
        return {'FINISHED'}


def register():
    bpy.utils.register_class(UvOperator)


def unregister():
    bpy.utils.unregister_class(UvOperator)


if __name__ == "__main__":
    register()

    # test call
    bpy.ops.uv.simple_operator()



# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

import bpy
import math

from collections import namedtuple

from bpy.props import (
    BoolProperty,
    IntProperty,
    FloatProperty,
    PointerProperty
)


ScatterSettings = namedtuple("ScatterSettings",
    ["seed", "density", "radius", "scale", "random_scale",
     "rotation", "normal_offset", "use_normal_rotation"])

class ObjectScatterProperties(bpy.types.PropertyGroup):
    seed: IntProperty(
        name="Seed",
        default=0,
        description="Change it to get a different scatter pattern",
    )

    density: FloatProperty(
        name="Density",
        default=10,
        min=0,
        soft_max=50,
        description="The amount of objects per unit on the line",
    )

    radius: FloatProperty(
        name="Radius",
        default=1,
        min=0,
        soft_max=5,
        subtype='DISTANCE',
        unit='LENGTH',
        description="Maximum distance of the objects to the line",
    )

    scale: FloatProperty(
        name="Scale",
        default=0.3,
        min=0.00001,
        soft_max=1,
        description="Size of the generated objects",
    )

    random_scale_percentage: FloatProperty(
        name="Random Scale Percentage",
        default=80,
        min=0,
        max=100,
        subtype='PERCENTAGE',
        precision=0,
        description="Increase to get a larger range of sizes",
    )

    rotation: FloatProperty(
        name="Rotation",
        default=0.5,
        min=0,
        max=math.pi * 2,
        soft_max=math.pi / 2,
        subtype='ANGLE',
        unit='ROTATION',
        description="Maximum rotation of the generated objects",
    )

    normal_offset: FloatProperty(
        name="Normal Offset",
        default=0,
        soft_min=-1.0,
        soft_max=1.0,
        description="Distance from the surface",
    )

    use_normal_rotation: BoolProperty(
        name="Use Normal Rotation",
        default=True,
        description="Rotate the instances according to the surface normals",
    )

    def to_settings(self):
        return ScatterSettings(
            seed=self.seed,
            density=self.density,
            radius=self.radius,
            scale=self.scale,
            random_scale=self.random_scale_percentage / 100,
            rotation=self.rotation,
            normal_offset=self.normal_offset,
            use_normal_rotation=self.use_normal_rotation,
        )


class ObjectScatterPanel(bpy.types.Panel):
    bl_idname = "OBJECT_PT_object_scatter"
    bl_label = "Object Scatter"
    bl_space_type = 'VIEW_3D'
    bl_region_type = 'UI'
    bl_category = "Tool"

    def draw(self, context):
        layout = self.layout
        layout.use_property_split = True
        scatter = context.scene.scatter_properties

        layout.prop(scatter, "density", slider=True)
        layout.prop(scatter, "radius", slider=True)

        col = layout.column(align=True)
        col.prop(scatter, "scale", slider=True)
        col.prop(scatter, "random_scale_percentage", text="Randomness", slider=True)

        layout.prop(scatter, "use_normal_rotation")
        layout.prop(scatter, "rotation", slider=True)
        layout.prop(scatter, "normal_offset", text="Offset", slider=True)
        layout.prop(scatter, "seed")

def draw_menu(self, context):
    layout = self.layout
    layout.operator("object.scatter")

classes = (
    ObjectScatterProperties,
    ObjectScatterPanel,
)

def register():
    for cls in classes:
        bpy.utils.register_class(cls)
    bpy.types.Scene.scatter_properties = PointerProperty(type=ObjectScatterProperties)
    bpy.types.VIEW3D_MT_object.append(draw_menu)

def unregister():
    for cls in classes:
        bpy.utils.unregister_class(cls)
    del bpy.types.Scene.scatter_properties
    bpy.types.VIEW3D_MT_object.remove(draw_menu)



# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####


from blenderkit import utils

import bpy
from bpy.app.handlers import persistent

import queue
import logging
bk_logger = logging.getLogger('blenderkit')

@persistent
def scene_load(context):
    user_preferences = bpy.context.preferences.addons['blenderkit'].preferences
    if user_preferences.use_timers:
        if not (bpy.app.timers.is_registered(queue_worker)):
            bpy.app.timers.register(queue_worker)


def get_queue():
    # we pick just a random one of blender types, to try to get a persistent queue
    t = bpy.types.Scene

    if not hasattr(t, 'task_queue'):
        t.task_queue = queue.Queue()
    return t.task_queue

class task_object:
    def __init__(self, command = '', arguments = (), wait = 0, only_last = False, fake_context = False, fake_context_area = 'VIEW_3D'):
        self.command = command
        self.arguments = arguments
        self.wait = wait
        self.only_last = only_last
        self.fake_context = fake_context
        self.fake_context_area = fake_context_area

def add_task(task, wait = 0, only_last = False, fake_context = False, fake_context_area = 'VIEW_3D'):
    q = get_queue()
    taskob = task_object(task[0],task[1], wait = wait, only_last = only_last, fake_context = fake_context, fake_context_area = fake_context_area)
    q.put(taskob)


def queue_worker():
    # utils.p('start queue worker timer')

    #bk_logger.debug('timer queue worker')
    time_step = 2.0
    q = get_queue()

    back_to_queue = [] #delayed events
    stashed = {}
    # first round we get all tasks that are supposed to be stashed and run only once (only_last option)
    # stashing finds tasks with the property only_last and same command and executes only the last one.
    while not q.empty():
        # print('queue while 1')

        task = q.get()
        if task.only_last:
            #this now makes the keys not only by task, but also first argument.
            # by now stashing is only used for ratings, where the first argument is url.
            # This enables fast rating of multiple assets while allowing larger delay for uploading of ratings.
            # this avoids a duplicate request error on the server
            stashed[str(task.command)+str(task.arguments[0])] = task
        else:
            back_to_queue.append(task)
    if len(stashed.keys())>1:
        bk_logger.debug('task queue stashed task:' +str(stashed))
    #return tasks to que except for stashed
    for task in back_to_queue:
        q.put(task)
    #return stashed tasks to queue
    for k in stashed.keys():
        q.put(stashed[k])
    #second round, execute or put back waiting tasks.
    back_to_queue = []
    while not q.empty():
        # print('window manager', bpy.context.window_manager)
        task = q.get()

        if task.wait>0:
            task.wait-=time_step
            back_to_queue.append(task)
        else:
            bk_logger.debug('task queue task:'+ str( task.command) +str( task.arguments))
            try:
                if task.fake_context:
                    fc = utils.get_fake_context(bpy.context, area_type = task.fake_context_area)
                    task.command(fc,*task.arguments)
                else:
                    task.command(*task.arguments)
            except Exception as e:
                bk_logger.error('task queue failed task:'+ str(task.command)+str(task.arguments)+ str(e))
                # bk_logger.exception('Got exception on main handler')
                # raise
        # print('queue while 2')
    for task in back_to_queue:
        q.put(task)
    # utils.p('end queue worker timer')

    return 2.0


def register():
    bpy.app.handlers.load_post.append(scene_load)


def unregister():
    bpy.app.handlers.load_post.remove(scene_load)



# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####
# Contributed to by:
# Pontiac, Fourmadmen, varkenvarken, tuga3d, meta-androcto, metalliandy     #
# dreampainter, cotejrp1, liero, Kayo Phoenix, sugiany, dommetysk, Jambay   #
# Phymec, Anthony D'Agostino, Pablo Vazquez, Richard Wilks, lijenstina,     #
# Sjaak-de-Draak, Phil Cote, cotejrp1, xyz presets by elfnor, revolt_randy, #
# Vladimir Spivak (cwolf3d), #


bl_info = {
    "name": "Extra Objects",
    "author": "Multiple Authors",
    "version": (0, 3, 7),
    "blender": (2, 80, 0),
    "location": "View3D > Add > Mesh",
    "description": "Add extra mesh object types",
    "warning": "",
    "doc_url": "{BLENDER_MANUAL_URL}/addons/add_mesh/mesh_extra_objects.html",
    "category": "Add Mesh",
}

# Note: Blocks has to be loaded before the WallFactory or the script
#       will not work properly after (F8) reload

if "bpy" in locals():
    import importlib
    importlib.reload(add_mesh_star)
    importlib.reload(add_mesh_twisted_torus)
    importlib.reload(add_mesh_gemstones)
    importlib.reload(add_mesh_gears)
    importlib.reload(add_mesh_3d_function_surface)
    importlib.reload(add_mesh_round_cube)
    importlib.reload(add_mesh_supertoroid)
    importlib.reload(add_mesh_pyramid)
    importlib.reload(add_mesh_torusknot)
    importlib.reload(add_mesh_honeycomb)
    importlib.reload(add_mesh_teapot)
    importlib.reload(add_mesh_pipe_joint)
    importlib.reload(add_mesh_solid)
    importlib.reload(add_mesh_round_brilliant)
    importlib.reload(add_mesh_menger_sponge)
    importlib.reload(add_mesh_vertex)
    importlib.reload(add_empty_as_parent)
    importlib.reload(add_mesh_beam_builder)
    importlib.reload(Blocks)
    importlib.reload(Wallfactory)
    importlib.reload(add_mesh_triangles)
else:
    from . import add_mesh_star
    from . import add_mesh_twisted_torus
    from . import add_mesh_gemstones
    from . import add_mesh_gears
    from . import add_mesh_3d_function_surface
    from . import add_mesh_round_cube
    from . import add_mesh_supertoroid
    from . import add_mesh_pyramid
    from . import add_mesh_torusknot
    from . import add_mesh_honeycomb
    from . import add_mesh_teapot
    from . import add_mesh_pipe_joint
    from . import add_mesh_solid
    from . import add_mesh_round_brilliant
    from . import add_mesh_menger_sponge
    from . import add_mesh_vertex
    from . import add_empty_as_parent
    from . import add_mesh_beam_builder
    from . import Blocks
    from . import Wallfactory
    from . import add_mesh_triangles

    from .add_mesh_rocks import __init__
    from .add_mesh_rocks import rockgen

import bpy
from bpy.types import Menu

class VIEW3D_MT_mesh_vert_add(Menu):
    # Define the "Single Vert" menu
    bl_idname = "VIEW3D_MT_mesh_vert_add"
    bl_label = "Single Vert"

    def draw(self, context):
        layout = self.layout
        layout.operator_context = 'INVOKE_REGION_WIN'
        layout.operator("mesh.primitive_vert_add",
                        text="Add Single Vert")
        layout.separator()
        layout.operator("mesh.primitive_emptyvert_add",
                        text="Object Origin Only")
        layout.operator("mesh.primitive_symmetrical_vert_add",
                        text="Origin & Vert Mirrored")
        layout.operator("mesh.primitive_symmetrical_empty_add",
                        text="Object Origin Mirrored")


class VIEW3D_MT_mesh_gears_add(Menu):
    # Define the "Gears" menu
    bl_idname = "VIEW3D_MT_mesh_gears_add"
    bl_label = "Gears"

    def draw(self, context):
        layout = self.layout
        layout.operator_context = 'INVOKE_REGION_WIN'
        oper = layout.operator("mesh.primitive_gear", text="Gear")
        oper.change = False
        oper = layout.operator("mesh.primitive_worm_gear", text="Worm")
        oper.change = False


class VIEW3D_MT_mesh_diamonds_add(Menu):
    # Define the "Diamonds" menu
    bl_idname = "VIEW3D_MT_mesh_diamonds_add"
    bl_label = "Diamonds"

    def draw(self, context):
        layout = self.layout
        layout.operator_context = 'INVOKE_REGION_WIN'
        oper = layout.operator("mesh.primitive_brilliant_add", text="Brilliant Diamond")
        oper.change = False
        oper = layout.operator("mesh.primitive_diamond_add", text="Diamond")
        oper.change = False
        oper = layout.operator("mesh.primitive_gem_add", text="Gem")
        oper.change = False


class VIEW3D_MT_mesh_math_add(Menu):
    # Define the "Math Function" menu
    bl_idname = "VIEW3D_MT_mesh_math_add"
    bl_label = "Math Functions"

    def draw(self, context):
        layout = self.layout
        layout.operator_context = 'INVOKE_REGION_WIN'
        layout.operator("mesh.primitive_z_function_surface",
                        text="Z Math Surface")
        layout.operator("mesh.primitive_xyz_function_surface",
                        text="XYZ Math Surface")
        self.layout.operator("mesh.primitive_solid_add", text="Regular Solid")
        self.layout.operator("mesh.make_triangle")


class VIEW3D_MT_mesh_extras_add(Menu):
    # Define the "Extra Objects" menu
    bl_idname = "VIEW3D_MT_mesh_extras_add"
    bl_label = "Extras"

    def draw(self, context):
        layout = self.layout
        layout.operator_context = 'INVOKE_REGION_WIN'
        oper = layout.operator("mesh.add_beam", text="Beam Builder")
        oper.change = False
        oper = layout.operator("mesh.wall_add", text="Wall Factory")
        oper.change = False
        layout.separator()
        oper = layout.operator("mesh.primitive_star_add", text="Simple Star")
        oper.change = False
        oper = layout.operator("mesh.primitive_steppyramid_add", text="Step Pyramid")
        oper.change = False
        oper = layout.operator("mesh.honeycomb_add", text="Honeycomb")
        oper.change = False
        oper = layout.operator("mesh.primitive_teapot_add", text="Teapot+")
        oper = layout.operator("mesh.menger_sponge_add", text="Menger Sponge")


class VIEW3D_MT_mesh_torus_add(Menu):
    # Define the "Torus Objects" menu
    bl_idname = "VIEW3D_MT_mesh_torus_add"
    bl_label = "Torus Objects"

    def draw(self, context):
        layout = self.layout
        layout.operator_context = 'INVOKE_REGION_WIN'
        oper = layout.operator("mesh.primitive_twisted_torus_add", text="Twisted Torus")
        oper.change = False
        oper = layout.operator("mesh.primitive_supertoroid_add", text="Supertoroid")
        oper.change = False
        oper = layout.operator("mesh.primitive_torusknot_add", text="Torus Knot")
        oper.change = False


class VIEW3D_MT_mesh_pipe_joints_add(Menu):
    # Define the "Pipe Joints" menu
    bl_idname = "VIEW3D_MT_mesh_pipe_joints_add"
    bl_label = "Pipe Joints"

    def draw(self, context):
        layout = self.layout
        layout.operator_context = 'INVOKE_REGION_WIN'
        oper = layout.operator("mesh.primitive_elbow_joint_add", text="Pipe Elbow")
        oper.change = False
        oper = layout.operator("mesh.primitive_tee_joint_add", text="Pipe T-Joint")
        oper.change = False
        oper = layout.operator("mesh.primitive_wye_joint_add", text="Pipe Y-Joint")
        oper.change = False
        oper = layout.operator("mesh.primitive_cross_joint_add", text="Pipe Cross-Joint")
        oper.change = False
        oper = layout.operator("mesh.primitive_n_joint_add", text="Pipe N-Joint")
        oper.change = False

# Register all operators and panels

# Define "Extras" menu
def menu_func(self, context):
    layout = self.layout
    layout.operator_context = 'INVOKE_REGION_WIN'

    layout.separator()
    layout.menu("VIEW3D_MT_mesh_vert_add",
                text="Single Vert", icon="DECORATE")
    oper = layout.operator("mesh.primitive_round_cube_add",
                    text="Round Cube", icon="SPHERE")
    oper.change = False
    layout.menu("VIEW3D_MT_mesh_torus_add",
                text="Torus Objects", icon="MESH_TORUS")
    layout.separator()
    layout.menu("VIEW3D_MT_mesh_math_add",
                text="Math Function", icon="PACKAGE")
    layout.menu("VIEW3D_MT_mesh_gears_add",
                text="Gears", icon="PREFERENCES")
    layout.menu("VIEW3D_MT_mesh_pipe_joints_add",
                text="Pipe Joints", icon="EMPTY_DATA")
    layout.separator()
    layout.menu("VIEW3D_MT_mesh_diamonds_add", text="Diamonds")
    layout.menu("VIEW3D_MT_mesh_extras_add",
                text="Extras")
    layout.separator()
    layout.operator("object.parent_to_empty",
                    text="Parent To Empty")

def Extras_contex_menu(self, context):
    bl_label = 'Change'

    obj = context.object
    layout = self.layout

    if obj == None or obj.data is None:
        return

    if 'Gear' in obj.data.keys():
        props = layout.operator("mesh.primitive_gear", text="Change Gear")
        props.change = True
        for prm in add_mesh_gears.GearParameters():
            setattr(props, prm, obj.data[prm])
        layout.separator()

    if 'WormGear' in obj.data.keys():
        props = layout.operator("mesh.primitive_worm_gear", text="Change WormGear")
        props.change = True
        for prm in add_mesh_gears.WormGearParameters():
            setattr(props, prm, obj.data[prm])
        layout.separator()

    if 'Beam' in obj.data.keys():
        props = layout.operator("mesh.add_beam", text="Change Beam")
        props.change = True
        for prm in add_mesh_beam_builder.BeamParameters():
            setattr(props, prm, obj.data[prm])
        layout.separator()

    if 'Wall' in obj.data.keys():
        props = layout.operator("mesh.wall_add", text="Change Wall")
        props.change = True
        for prm in Wallfactory.WallParameters():
            setattr(props, prm, obj.data[prm])
        layout.separator()

    if 'ElbowJoint' in obj.data.keys():
        props = layout.operator("mesh.primitive_elbow_joint_add", text="Change ElbowJoint")
        props.change = True
        for prm in add_mesh_pipe_joint.ElbowJointParameters():
            setattr(props, prm, obj.data[prm])
        layout.separator()

    if 'TeeJoint' in obj.data.keys():
        props = layout.operator("mesh.primitive_tee_joint_add", text="Change TeeJoint")
        props.change = True
        for prm in add_mesh_pipe_joint.TeeJointParameters():
            setattr(props, prm, obj.data[prm])
        layout.separator()

    if 'WyeJoint' in obj.data.keys():
        props = layout.operator("mesh.primitive_wye_joint_add", text="Change WyeJoint")
        props.change = True
        for prm in add_mesh_pipe_joint.WyeJointParameters():
            setattr(props, prm, obj.data[prm])
        layout.separator()

    if 'CrossJoint' in obj.data.keys():
        props = layout.operator("mesh.primitive_cross_joint_add", text="Change CrossJoint")
        props.change = True
        for prm in add_mesh_pipe_joint.CrossJointParameters():
            setattr(props, prm, obj.data[prm])
        layout.separator()

    if 'NJoint' in obj.data.keys():
        props = layout.operator("mesh.primitive_n_joint_add", text="Change NJoint")
        props.change = True
        for prm in add_mesh_pipe_joint.NJointParameters():
            setattr(props, prm, obj.data[prm])
        layout.separator()

    if 'Diamond' in obj.data.keys():
        props = layout.operator("mesh.primitive_diamond_add", text="Change Diamond")
        props.change = True
        for prm in add_mesh_gemstones.DiamondParameters():
            setattr(props, prm, obj.data[prm])
        layout.separator()

    if 'Gem' in obj.data.keys():
        props = layout.operator("mesh.primitive_gem_add", text="Change Gem")
        props.change = True
        for prm in add_mesh_gemstones.GemParameters():
            setattr(props, prm, obj.data[prm])
        layout.separator()

    if 'Brilliant' in obj.data.keys():
        props = layout.operator("mesh.primitive_brilliant_add", text="Change Brilliant")
        props.change = True
        for prm in add_mesh_round_brilliant.BrilliantParameters():
            setattr(props, prm, obj.data[prm])
        layout.separator()

    if 'Roundcube' in obj.data.keys():
        props = layout.operator("mesh.primitive_round_cube_add", text="Change Roundcube")
        props.change = True
        for prm in add_mesh_round_cube.RoundCubeParameters():
            setattr(props, prm, obj.data[prm])
        layout.separator()

    if 'TorusKnot' in obj.data.keys():
        props = layout.operator("mesh.primitive_torusknot_add", text="Change TorusKnot")
        props.change = True
        for prm in add_mesh_torusknot.TorusKnotParameters():
            setattr(props, prm, obj.data[prm])
        layout.separator()

    if 'SuperToroid' in obj.data.keys():
        props = layout.operator("mesh.primitive_supertoroid_add", text="Change SuperToroid")
        props.change = True
        for prm in add_mesh_supertoroid.SuperToroidParameters():
            setattr(props, prm, obj.data[prm])
        layout.separator()

    if 'TwistedTorus' in obj.data.keys():
        props = layout.operator("mesh.primitive_twisted_torus_add", text="Change TwistedTorus")
        props.change = True
        for prm in add_mesh_twisted_torus.TwistedTorusParameters():
            setattr(props, prm, obj.data[prm])
        layout.separator()

    if 'Star' in obj.data.keys():
        props = layout.operator("mesh.primitive_star_add", text="Change Star")
        props.change = True
        for prm in add_mesh_star.StarParameters():
            setattr(props, prm, obj.data[prm])
        layout.separator()

    if 'Pyramid' in obj.data.keys():
        props = layout.operator("mesh.primitive_steppyramid_add", text="Change Pyramid")
        props.change = True
        for prm in add_mesh_pyramid.PyramidParameters():
            setattr(props, prm, obj.data[prm])
        layout.separator()

    if 'HoneyComb' in obj.data.keys():
        props = layout.operator("mesh.honeycomb_add", text="Change HoneyComb")
        props.change = True
        for prm in add_mesh_honeycomb.HoneyCombParameters():
            setattr(props, prm, obj.data[prm])
        layout.separator()

# Register
classes = [
    VIEW3D_MT_mesh_vert_add,
    VIEW3D_MT_mesh_gears_add,
    VIEW3D_MT_mesh_diamonds_add,
    VIEW3D_MT_mesh_math_add,
    VIEW3D_MT_mesh_extras_add,
    VIEW3D_MT_mesh_torus_add,
    VIEW3D_MT_mesh_pipe_joints_add,
    add_mesh_star.AddStar,
    add_mesh_twisted_torus.AddTwistedTorus,
    add_mesh_gemstones.AddDiamond,
    add_mesh_gemstones.AddGem,
    add_mesh_gears.AddGear,
    add_mesh_gears.AddWormGear,
    add_mesh_3d_function_surface.AddZFunctionSurface,
    add_mesh_3d_function_surface.AddXYZFunctionSurface,
    add_mesh_round_cube.AddRoundCube,
    add_mesh_supertoroid.add_supertoroid,
    add_mesh_pyramid.AddPyramid,
    add_mesh_torusknot.AddTorusKnot,
    add_mesh_honeycomb.add_mesh_honeycomb,
    add_mesh_teapot.AddTeapot,
    add_mesh_pipe_joint.AddElbowJoint,
    add_mesh_pipe_joint.AddTeeJoint,
    add_mesh_pipe_joint.AddWyeJoint,
    add_mesh_pipe_joint.AddCrossJoint,
    add_mesh_pipe_joint.AddNJoint,
    add_mesh_solid.Solids,
    add_mesh_round_brilliant.MESH_OT_primitive_brilliant_add,
    add_mesh_menger_sponge.AddMengerSponge,
    add_mesh_vertex.AddVert,
    add_mesh_vertex.AddEmptyVert,
    add_mesh_vertex.AddSymmetricalEmpty,
    add_mesh_vertex.AddSymmetricalVert,
    add_empty_as_parent.P2E,
    add_empty_as_parent.PreFix,
    add_mesh_beam_builder.addBeam,
    Wallfactory.add_mesh_wallb,
    add_mesh_triangles.MakeTriangle,
]

def register():
    from bpy.utils import register_class
    for cls in classes:
        register_class(cls)

    add_mesh_rocks.register()

    # Add "Extras" menu to the "Add Mesh" menu and context menu.
    bpy.types.VIEW3D_MT_mesh_add.append(menu_func)
    bpy.types.VIEW3D_MT_object_context_menu.prepend(Extras_contex_menu)


def unregister():
    # Remove "Extras" menu from the "Add Mesh" menu and context menu.
    bpy.types.VIEW3D_MT_object_context_menu.remove(Extras_contex_menu)
    bpy.types.VIEW3D_MT_mesh_add.remove(menu_func)

    from bpy.utils import unregister_class
    for cls in reversed(classes):
        unregister_class(cls)

    add_mesh_rocks.unregister()

if __name__ == "__main__":
    register()



# -*- coding:utf-8 -*-

# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110- 1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

# <pep8 compliant>

# ----------------------------------------------------------
# Author: Stephen Leger (s-leger)
# Inspired by Asset-Flinguer
# ----------------------------------------------------------
import sys
from mathutils import Vector
import bpy


def log(s):
    print("[log]" + s)


def create_lamp(context, loc):
    bpy.ops.object.light_add(
        type='POINT',
        radius=1,
        align='WORLD',
        location=loc)
    lamp = context.active_object
    lamp.data.use_nodes = True
    lamp.location = loc
    tree = lamp.data.node_tree
    return tree, tree.nodes, lamp.data


def create_camera(context, loc, rot):
    bpy.ops.object.camera_add(
        align='WORLD',
        enter_editmode=False,
        location=loc,
        rotation=rot)
    cam = context.active_object
    cam.location = loc
    cam.rotation_euler = rot
    context.scene.camera = cam
    return cam


def get_center(o):
    x, y, z = o.bound_box[0]
    min_x = x
    min_y = y
    min_z = z
    x, y, z = o.bound_box[6]
    max_x = x
    max_y = y
    max_z = z
    return Vector((
        min_x + 0.5 * (max_x - min_x),
        min_y + 0.5 * (max_y - min_y),
        min_z + 0.5 * (max_z - min_z)))


def apply_simple_material(o, name, color):
    m = bpy.data.materials.new(name)
    m.use_nodes = True
    for node in m.node_tree.nodes:
        if node.bl_idname == "ShaderNodeBsdfPrincipled":
            node.inputs[0].default_value = color
            break
    o.data.materials.append(m)


def generateThumb(context, cls, preset, engine):
    log("### RENDER THUMB ############################")

    # Cleanup scene
    for o in context.scene.objects:
        o.select_set(state=True)

    bpy.ops.object.delete()

    log("Start generating: %s" % cls)

    # setup render

    context.scene.render.engine = engine

    if engine == 'CYCLES':
        cycles = context.scene.cycles
        cycles.progressive = 'PATH'
        cycles.samples = 24
        try:
            cycles.use_square_samples = True
        except:
            pass
        cycles.preview_samples = 24
        cycles.aa_samples = 24
        cycles.transparent_max_bounces = 8
        cycles.transparent_min_bounces = 8
        cycles.transmission_bounces = 8
        cycles.max_bounces = 8
        cycles.min_bounces = 6
        cycles.caustics_refractive = False
        cycles.caustics_reflective = False
        cycles.use_transparent_shadows = True
        cycles.diffuse_bounces = 1
        cycles.glossy_bounces = 4

    elif engine == 'BLENDER_EEVEE':
        eevee = context.scene.eevee
        eevee.use_gtao = True
        eevee.use_ssr = True
        eevee.use_soft_shadows = True
        eevee.taa_render_samples = 64
    else:
        raise RuntimeError("Unsupported render engine %s" % engine)

    render = context.scene.render

    # engine settings
    render.resolution_x = 150
    render.resolution_y = 100
    render.filepath = preset[:-3] + ".png"

    # create object, loading preset
    getattr(bpy.ops.archipack, cls)('INVOKE_DEFAULT', filepath=preset, auto_manipulate=False)
    o = context.active_object
    size = o.dimensions
    center = get_center(o)

    # opposite / tan (0.5 * fov)  where fov is 49.134 deg
    dist = max(size) / 0.32
    loc = center + dist * Vector((-0.5, -1, 0.5)).normalized()

    log("Prepare camera")
    cam = create_camera(context, loc, (1.150952, 0.0, -0.462509))
    cam.data.lens = 50

    for ob in context.scene.objects:
        ob.select_set(state=False)

    o.select_set(state=True)

    bpy.ops.view3d.camera_to_view_selected()
    cam.data.lens = 45

    log("Prepare scene")
    # add plane
    bpy.ops.mesh.primitive_plane_add(
        size=1000,
        align='WORLD',
        enter_editmode=False,
        location=(0, 0, 0)
    )

    p = context.active_object
    apply_simple_material(p, "Plane", (1, 1, 1, 1))

    # add 3 lights
    tree, nodes, lamp = create_lamp(context, (3.69736, -7, 6.0))
    lamp.energy = 1000
    emit = nodes["Emission"]
    emit.inputs[1].default_value = 2000.0

    tree, nodes, lamp = create_lamp(context, (9.414563179016113, 5.446230888366699, 5.903861999511719))
    lamp.energy = 400
    emit = nodes["Emission"]
    falloff = nodes.new(type="ShaderNodeLightFalloff")
    falloff.inputs[0].default_value = 5
    tree.links.new(falloff.outputs[2], emit.inputs[1])

    tree, nodes, lamp = create_lamp(context, (-7.847615718841553, 1.03135085105896, 5.903861999511719))
    lamp.energy = 200
    emit = nodes["Emission"]
    falloff = nodes.new(type="ShaderNodeLightFalloff")
    falloff.inputs[0].default_value = 5
    tree.links.new(falloff.outputs[2], emit.inputs[1])

    # Set output filename.
    render.use_file_extension = True
    render.use_overwrite = True
    render.use_compositing = False
    render.use_sequencer = False
    render.resolution_percentage = 100
    # render.image_settings.file_format = 'PNG'
    # render.image_settings.color_mode = 'RGBA'
    # render.image_settings.color_depth = '8'

    # Configure output size.
    log("Render")

    # Render thumbnail
    bpy.ops.render.render(write_still=True)

    log("### COMPLETED ############################")


if __name__ == "__main__":

    preset = ""
    engine = 'BLENDER_EEVEE' #'CYCLES'
    for arg in sys.argv:
        if arg.startswith("cls:"):
            cls = arg[4:]
        if arg.startswith("preset:"):
            preset = arg[7:]
        if arg.startswith("matlib:"):
            matlib = arg[7:]
        if arg.startswith("addon:"):
            module = arg[6:]
        if arg.startswith("engine:"):
            engine = arg[7:]
    try:
        # log("### ENABLE %s ADDON ############################" % module)
        bpy.ops.preferences.addon_enable(module=module)
        # log("### MATLIB PATH ############################")
        bpy.context.preferences.addons[module].preferences.matlib_path = matlib
    except:
        raise RuntimeError("module name not found")
    # log("### GENERATE ############################")
    generateThumb(bpy.context, cls, preset, engine)



# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

# <pep8 compliant>

bl_info = {
    "name": "Hotkey: 'X'",
    "description": "Edit mode V/E/F Delete Modes",
    "author": "pitiwazou, meta-androcto",
    "version": (0, 1, 0),
    "blender": (2, 80, 0),
    "location": "Mesh Edit Mode",
    "warning": "",
    "doc_url": "",
    "category": "Edit Delete Pie"
    }

import bpy
from bpy.types import Menu


# Pie Delete - X
class PIE_MT_PieDelete(Menu):
    bl_idname = "PIE_MT_delete"
    bl_label = "Pie Delete"

    def draw(self, context):
        layout = self.layout
        pie = layout.menu_pie()
        # 4 - LEFT
        box = pie.split().column()
        box.operator("mesh.dissolve_limited", text="Limited Dissolve", icon='STICKY_UVS_LOC')
        box.operator("mesh.delete_edgeloop", text="Delete Edge Loops", icon='NONE')
        box.operator("mesh.edge_collapse", text="Edge Collapse", icon='UV_EDGESEL')
        # 6 - RIGHT
        box = pie.split().column()
        box.operator("mesh.remove_doubles", text="Merge By Distance", icon='NONE')
        box.operator("mesh.delete", text="Only Edge & Faces", icon='NONE').type = 'EDGE_FACE'
        box.operator("mesh.delete", text="Only Faces", icon='UV_FACESEL').type = 'ONLY_FACE'
        # 2 - BOTTOM
        pie.operator("mesh.dissolve_edges", text="Dissolve Edges", icon='SNAP_EDGE')
        # 8 - TOP
        pie.operator("mesh.delete", text="Delete Edges", icon='EDGESEL').type = 'EDGE'
        # 7 - TOP - LEFT
        pie.operator("mesh.delete", text="Delete Vertices", icon='VERTEXSEL').type = 'VERT'
        # 9 - TOP - RIGHT
        pie.operator("mesh.delete", text="Delete Faces", icon='FACESEL').type = 'FACE'
        # 1 - BOTTOM - LEFT
        pie.operator("mesh.dissolve_verts", text="Dissolve Vertices", icon='SNAP_VERTEX')
        # 3 - BOTTOM - RIGHT
        pie.operator("mesh.dissolve_faces", text="Dissolve Faces", icon='SNAP_FACE')


classes = (
    PIE_MT_PieDelete,
    )


addon_keymaps = []


def register():
    for cls in classes:
        bpy.utils.register_class(cls)

    wm = bpy.context.window_manager
    if wm.keyconfigs.addon:
        # Delete
        km = wm.keyconfigs.addon.keymaps.new(name='Mesh')
        kmi = km.keymap_items.new('wm.call_menu_pie', 'X', 'PRESS')
        kmi.properties.name = "PIE_MT_delete"
        addon_keymaps.append((km, kmi))


def unregister():
    for cls in classes:
        bpy.utils.unregister_class(cls)

    wm = bpy.context.window_manager
    kc = wm.keyconfigs.addon
    if kc:
        for km, kmi in addon_keymaps:
            km.keymap_items.remove(kmi)
    addon_keymaps.clear()


if __name__ == "__main__":
    register()



#!/usr/bin/env python3

# ***** BEGIN GPL LICENSE BLOCK *****
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ***** END GPL LICENSE BLOCK *****

# <pep8 compliant>

CHECKER_IGNORE_PREFIX = [
    "extern",
    "intern/moto",
]

CHECKER_BIN = "sparse"
CHECKER_ARGS = [
]

import project_source_info
import subprocess
import sys
import os

USE_QUIET = (os.environ.get("QUIET", None) is not None)


def main():
    source_info = project_source_info.build_info(use_cxx=False, ignore_prefix_list=CHECKER_IGNORE_PREFIX)
    source_defines = project_source_info.build_defines_as_args()

    check_commands = []
    for c, inc_dirs, defs in source_info:

        cmd = ([CHECKER_BIN] +
               CHECKER_ARGS +
               [c] +
               [("-I%s" % i) for i in inc_dirs] +
               [("-D%s" % d) for d in defs] +
               source_defines
               )

        check_commands.append((c, cmd))

    def my_process(i, c, cmd):
        if not USE_QUIET:
            percent = 100.0 * (i / len(check_commands))
            percent_str = "[" + ("%.2f]" % percent).rjust(7) + " %:"

            sys.stdout.flush()
            sys.stdout.write("%s %s\n" % (percent_str, c))

        return subprocess.Popen(cmd)

    process_functions = []
    for i, (c, cmd) in enumerate(check_commands):
        process_functions.append((my_process, (i, c, cmd)))

    project_source_info.queue_processes(process_functions)


if __name__ == "__main__":
    main()



# <pep8-80 compliant>

# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

__author__ = "McBuff, Nutti <nutti.metro@gmail.com>"
__status__ = "production"
__version__ = "6.5"
__date__ = "6 Mar 2021"

from math import sqrt

import bpy
from bpy.props import (
    EnumProperty,
    FloatProperty,
    IntVectorProperty,
    BoolProperty,
)
import bmesh
from mathutils import Vector

from .. import common
from ..utils.bl_class_registry import BlClassRegistry
from ..utils.property_class_registry import PropertyClassRegistry
from ..utils import compatibility as compat


def _is_valid_context_for_measure(context):
    # Multiple objects editing mode is not supported in this feature.
    objs = common.get_uv_editable_objects(context)
    if len(objs) != 1:
        return False

    # only edit mode is allowed to execute
    if context.object.mode != 'EDIT':
        return False

    # only 'VIEW_3D' space is allowed to execute
    if not common.is_valid_space(context, ['VIEW_3D']):
        return False

    return True


def _is_valid_context_for_apply(context):
    objs = common.get_uv_editable_objects(context)
    if not objs:
        return False

    # only edit mode is allowed to execute
    if context.object.mode != 'EDIT':
        return False

    # only 'VIEW_3D' space is allowed to execute
    if not common.is_valid_space(context, ['VIEW_3D']):
        return False

    return True


def _measure_wsuv_info(obj, calc_method='MESH',
                       tex_selection_method='FIRST', tex_size=None,
                       only_selected=True):
    mesh_areas = common.measure_mesh_area(obj, calc_method, only_selected)
    uv_areas = common.measure_uv_area(obj, calc_method, tex_selection_method,
                                      tex_size, only_selected)

    if not uv_areas:
        return None, mesh_areas, None

    if len(mesh_areas) != len(uv_areas):
        raise ValueError("mesh_area and uv_area must be same length")

    densities = []
    for mesh_area, uv_area in zip(mesh_areas, uv_areas):
        if mesh_area == 0.0:
            densities.append(0.0)
        else:
            densities.append(sqrt(uv_area) / sqrt(mesh_area))

    return uv_areas, mesh_areas, densities


def _measure_wsuv_info_from_faces(obj, bm, faces, uv_layer, tex_layer,
                                  tex_selection_method='FIRST', tex_size=None):
    mesh_area = common.measure_mesh_area_from_faces(bm, faces)
    uv_area = common.measure_uv_area_from_faces(
        obj, bm, faces, uv_layer, tex_layer, tex_selection_method, tex_size)

    if not uv_area:
        return None, mesh_area, None

    if mesh_area == 0.0:
        density = 0.0
    else:
        density = sqrt(uv_area) / sqrt(mesh_area)

    return uv_area, mesh_area, density


def _apply(faces, uv_layer, origin, factor):
    # calculate origin
    if origin == 'CENTER':
        origin = Vector((0.0, 0.0))
        num = 0
        for f in faces:
            for l in f.loops:
                uv = l[uv_layer].uv
                origin = origin + uv
                num = num + 1
        origin = origin / num
    elif origin == 'LEFT_TOP':
        origin = Vector((100000.0, -100000.0))
        for f in faces:
            for l in f.loops:
                uv = l[uv_layer].uv
                origin.x = min(origin.x, uv.x)
                origin.y = max(origin.y, uv.y)
    elif origin == 'LEFT_CENTER':
        origin = Vector((100000.0, 0.0))
        num = 0
        for f in faces:
            for l in f.loops:
                uv = l[uv_layer].uv
                origin.x = min(origin.x, uv.x)
                origin.y = origin.y + uv.y
                num = num + 1
        origin.y = origin.y / num
    elif origin == 'LEFT_BOTTOM':
        origin = Vector((100000.0, 100000.0))
        for f in faces:
            for l in f.loops:
                uv = l[uv_layer].uv
                origin.x = min(origin.x, uv.x)
                origin.y = min(origin.y, uv.y)
    elif origin == 'CENTER_TOP':
        origin = Vector((0.0, -100000.0))
        num = 0
        for f in faces:
            for l in f.loops:
                uv = l[uv_layer].uv
                origin.x = origin.x + uv.x
                origin.y = max(origin.y, uv.y)
                num = num + 1
        origin.x = origin.x / num
    elif origin == 'CENTER_BOTTOM':
        origin = Vector((0.0, 100000.0))
        num = 0
        for f in faces:
            for l in f.loops:
                uv = l[uv_layer].uv
                origin.x = origin.x + uv.x
                origin.y = min(origin.y, uv.y)
                num = num + 1
        origin.x = origin.x / num
    elif origin == 'RIGHT_TOP':
        origin = Vector((-100000.0, -100000.0))
        for f in faces:
            for l in f.loops:
                uv = l[uv_layer].uv
                origin.x = max(origin.x, uv.x)
                origin.y = max(origin.y, uv.y)
    elif origin == 'RIGHT_CENTER':
        origin = Vector((-100000.0, 0.0))
        num = 0
        for f in faces:
            for l in f.loops:
                uv = l[uv_layer].uv
                origin.x = max(origin.x, uv.x)
                origin.y = origin.y + uv.y
                num = num + 1
        origin.y = origin.y / num
    elif origin == 'RIGHT_BOTTOM':
        origin = Vector((-100000.0, 100000.0))
        for f in faces:
            for l in f.loops:
                uv = l[uv_layer].uv
                origin.x = max(origin.x, uv.x)
                origin.y = min(origin.y, uv.y)

    # update UV coordinate
    for f in faces:
        for l in f.loops:
            uv = l[uv_layer].uv
            diff = uv - origin
            l[uv_layer].uv = origin + diff * factor


def _get_target_textures(_, __):
    objs = common.get_uv_editable_objects(bpy.context)
    images = []
    for obj in objs:
        images.extend(common.find_images(obj))

    items = []
    items.append(("[Average]", "[Average]", "Average of all textures"))
    items.append(("[Max]", "[Max]", "Max of all textures"))
    items.append(("[Min]", "[Min]", "Min of all textures"))
    items.extend([(img.name, img.name, "") for img in images])
    return items


@PropertyClassRegistry()
class _Properties:
    idname = "world_scale_uv"

    @classmethod
    def init_props(cls, scene):
        scene.muv_world_scale_uv_enabled = BoolProperty(
            name="World Scale UV Enabled",
            description="World Scale UV is enabled",
            default=False
        )
        scene.muv_world_scale_uv_src_mesh_area = FloatProperty(
            name="Mesh Area",
            description="Source Mesh Area",
            default=0.0,
            min=0.0
        )
        scene.muv_world_scale_uv_src_uv_area = FloatProperty(
            name="UV Area",
            description="Source UV Area (Average if calculation method is UV "
                        "Island or Face)",
            default=0.0,
            min=0.0
        )
        scene.muv_world_scale_uv_src_density = FloatProperty(
            name="Density",
            description="Source Texel Density",
            default=0.0,
            min=0.0
        )
        scene.muv_world_scale_uv_tgt_density = FloatProperty(
            name="Density",
            description="Target Texel Density",
            default=0.0,
            min=0.0
        )
        scene.muv_world_scale_uv_tgt_scaling_factor = FloatProperty(
            name="Scaling Factor",
            default=1.0,
            max=1000.0,
            min=0.00001
        )
        scene.muv_world_scale_uv_tgt_texture_size = IntVectorProperty(
            name="Texture Size",
            size=2,
            min=1,
            soft_max=10240,
            default=(1024, 1024),
        )
        scene.muv_world_scale_uv_mode = EnumProperty(
            name="Mode",
            description="Density calculation mode",
            items=[
                ('PROPORTIONAL_TO_MESH', "Proportional to Mesh",
                 "Apply density proportionaled by mesh size"),
                ('SCALING_DENSITY', "Scaling Density",
                 "Apply scaled density from source"),
                ('SAME_DENSITY', "Same Density",
                 "Apply same density of source"),
                ('MANUAL', "Manual", "Specify density and size by manual"),
            ],
            default='MANUAL'
        )
        scene.muv_world_scale_uv_origin = EnumProperty(
            name="Origin",
            description="Aspect Origin",
            items=[
                ('CENTER', "Center", "Center"),
                ('LEFT_TOP', "Left Top", "Left Bottom"),
                ('LEFT_CENTER', "Left Center", "Left Center"),
                ('LEFT_BOTTOM', "Left Bottom", "Left Bottom"),
                ('CENTER_TOP', "Center Top", "Center Top"),
                ('CENTER_BOTTOM', "Center Bottom", "Center Bottom"),
                ('RIGHT_TOP', "Right Top", "Right Top"),
                ('RIGHT_CENTER', "Right Center", "Right Center"),
                ('RIGHT_BOTTOM', "Right Bottom", "Right Bottom")

            ],
            default='CENTER',
        )
        scene.muv_world_scale_uv_measure_tgt_texture = EnumProperty(
            name="Texture",
            description="Texture to be measured",
            items=_get_target_textures
        )
        scene.muv_world_scale_uv_apply_tgt_texture = EnumProperty(
            name="Texture",
            description="Texture to be applied",
            items=_get_target_textures
        )
        scene.muv_world_scale_uv_tgt_area_calc_method = EnumProperty(
            name="Area Calculation Method",
            description="How to calculate target area",
            items=[
                ('MESH', "Mesh", "Calculate area by whole faces in mesh"),
                ('UV ISLAND', "UV Island", "Calculate area each UV islands"),
                ('FACE', "Face", "Calculate area each face")
            ],
            default='MESH'
        )
        scene.muv_world_scale_uv_measure_only_selected = BoolProperty(
            name="Only Selected",
            description="Measure with only selected faces",
            default=True,
        )
        scene.muv_world_scale_uv_apply_only_selected = BoolProperty(
            name="Only Selected",
            description="Apply to only selected faces",
            default=True,
        )

    @classmethod
    def del_props(cls, scene):
        del scene.muv_world_scale_uv_enabled
        del scene.muv_world_scale_uv_src_mesh_area
        del scene.muv_world_scale_uv_src_uv_area
        del scene.muv_world_scale_uv_src_density
        del scene.muv_world_scale_uv_tgt_density
        del scene.muv_world_scale_uv_tgt_scaling_factor
        del scene.muv_world_scale_uv_mode
        del scene.muv_world_scale_uv_origin
        del scene.muv_world_scale_uv_measure_tgt_texture
        del scene.muv_world_scale_uv_apply_tgt_texture
        del scene.muv_world_scale_uv_tgt_area_calc_method
        del scene.muv_world_scale_uv_measure_only_selected
        del scene.muv_world_scale_uv_apply_only_selected


@BlClassRegistry()
@compat.make_annotations
class MUV_OT_WorldScaleUV_Measure(bpy.types.Operator):
    """
    Operation class: Measure face size
    """

    bl_idname = "uv.muv_world_scale_uv_measure"
    bl_label = "Measure World Scale UV"
    bl_description = "Measure face size for scale calculation"
    bl_options = {'REGISTER', 'UNDO'}

    tgt_texture = EnumProperty(
        name="Texture",
        description="Texture to be applied",
        items=_get_target_textures
    )
    only_selected = BoolProperty(
        name="Only Selected",
        description="Measure with only selected faces",
        default=True,
    )

    @classmethod
    def poll(cls, context):
        # we can not get area/space/region from console
        if common.is_console_mode():
            return True
        return _is_valid_context_for_measure(context)

    @staticmethod
    def setup_argument(ops, scene):
        try:
            ops.tgt_texture = scene.muv_world_scale_uv_measure_tgt_texture
        except TypeError:
            # Workaround for the error raised when the items of EnumProperty
            # are deleted.
            ops.tgt_texture = "[Average]"
        ops.only_selected = scene.muv_world_scale_uv_measure_only_selected

    def execute(self, context):
        sc = context.scene
        objs = common.get_uv_editable_objects(context)
        # poll() method ensures that only one object is selected.
        obj = objs[0]

        if self.tgt_texture == "[Average]":
            uv_areas, mesh_areas, densities = _measure_wsuv_info(
                obj, calc_method='MESH', tex_selection_method='AVERAGE',
                only_selected=self.only_selected)
        elif self.tgt_texture == "[Max]":
            uv_areas, mesh_areas, densities = _measure_wsuv_info(
                obj, calc_method='MESH', tex_selection_method='MAX',
                only_selected=self.only_selected)
        elif self.tgt_texture == "[Min]":
            uv_areas, mesh_areas, densities = _measure_wsuv_info(
                obj, calc_method='MESH', tex_selection_method='MIN',
                only_selected=self.only_selected)
        else:
            texture = bpy.data.images[self.tgt_texture]
            uv_areas, mesh_areas, densities = _measure_wsuv_info(
                obj, calc_method='MESH', tex_selection_method='USER_SPECIFIED',
                only_selected=self.only_selected, tex_size=texture.size)
        if not uv_areas:
            self.report({'WARNING'},
                        "Object must have more than one UV map and texture")
            return {'CANCELLED'}

        sc.muv_world_scale_uv_src_uv_area = uv_areas[0]
        sc.muv_world_scale_uv_src_mesh_area = mesh_areas[0]
        sc.muv_world_scale_uv_src_density = densities[0]

        self.report({'INFO'},
                    "UV Area: {0}, Mesh Area: {1}, Texel Density: {2}"
                    .format(uv_areas[0], mesh_areas[0], densities[0]))

        return {'FINISHED'}


@BlClassRegistry()
@compat.make_annotations
class MUV_OT_WorldScaleUV_ApplyManual(bpy.types.Operator):
    """
    Operation class: Apply scaled UV (Manual)
    """

    bl_idname = "uv.muv_world_scale_uv_apply_manual"
    bl_label = "Apply World Scale UV (Manual)"
    bl_description = "Apply scaled UV based on user specification"
    bl_options = {'REGISTER', 'UNDO'}

    tgt_density = FloatProperty(
        name="Density",
        description="Target Texel Density",
        default=1.0,
        min=0.0
    )
    tgt_texture_size = IntVectorProperty(
        name="Texture Size",
        size=2,
        min=1,
        soft_max=10240,
        default=(1024, 1024),
    )
    origin = EnumProperty(
        name="Origin",
        description="Aspect Origin",
        items=[
            ('CENTER', "Center", "Center"),
            ('LEFT_TOP', "Left Top", "Left Bottom"),
            ('LEFT_CENTER', "Left Center", "Left Center"),
            ('LEFT_BOTTOM', "Left Bottom", "Left Bottom"),
            ('CENTER_TOP', "Center Top", "Center Top"),
            ('CENTER_BOTTOM', "Center Bottom", "Center Bottom"),
            ('RIGHT_TOP', "Right Top", "Right Top"),
            ('RIGHT_CENTER', "Right Center", "Right Center"),
            ('RIGHT_BOTTOM', "Right Bottom", "Right Bottom")

        ],
        default='CENTER'
    )
    show_dialog = BoolProperty(
        name="Show Diaglog Menu",
        description="Show dialog menu if true",
        default=True,
        options={'HIDDEN', 'SKIP_SAVE'}
    )
    tgt_area_calc_method = EnumProperty(
        name="Area Calculation Method",
        description="How to calculate target area",
        items=[
            ('MESH', "Mesh", "Calculate area by whole faces in mesh"),
            ('UV ISLAND', "UV Island", "Calculate area each UV islands"),
            ('FACE', "Face", "Calculate area each face")
        ],
        default='MESH'
    )
    only_selected = BoolProperty(
        name="Only Selected",
        description="Apply to only selected faces",
        default=True,
    )

    @classmethod
    def poll(cls, context):
        # we can not get area/space/region from console
        if common.is_console_mode():
            return True
        return _is_valid_context_for_apply(context)

    @staticmethod
    def setup_argument(ops, scene):
        ops.tgt_density = scene.muv_world_scale_uv_tgt_density
        ops.tgt_texture_size = scene.muv_world_scale_uv_tgt_texture_size
        ops.origin = scene.muv_world_scale_uv_origin
        ops.show_dialog = False
        ops.tgt_area_calc_method = \
            scene.muv_world_scale_uv_tgt_area_calc_method
        ops.only_selected = scene.muv_world_scale_uv_apply_only_selected

    def __apply_manual(self, context):
        objs = common.get_uv_editable_objects(context)

        for obj in objs:
            bm = bmesh.from_edit_mesh(obj.data)
            if common.check_version(2, 73, 0) >= 0:
                bm.verts.ensure_lookup_table()
                bm.edges.ensure_lookup_table()
                bm.faces.ensure_lookup_table()

            if not bm.loops.layers.uv:
                self.report({'WARNING'},
                            "Object {} must have more than one UV map"
                            .format(obj.name))
                return {'CANCELLED'}
            uv_layer = bm.loops.layers.uv.verify()
            tex_layer = common.find_texture_layer(bm)
            faces_list = common.get_faces_list(
                bm, self.tgt_area_calc_method, self.only_selected)

            tex_size = self.tgt_texture_size

            factors = []
            for faces in faces_list:
                uv_area, _, density = _measure_wsuv_info_from_faces(
                    obj, bm, faces, uv_layer, tex_layer,
                    tex_selection_method='USER_SPECIFIED', tex_size=tex_size)

                if not uv_area:
                    self.report({'WARNING'},
                                "Object {} must have more than one UV map"
                                .format(obj.name))
                    return {'CANCELLED'}

                tgt_density = self.tgt_density
                factor = tgt_density / density

                _apply(faces, uv_layer, self.origin, factor)
                factors.append(factor)

            bmesh.update_edit_mesh(obj.data)
            self.report({'INFO'},
                        "Scaling factor of object {}: {}"
                        .format(obj.name, factors))

        return {'FINISHED'}

    def draw(self, _):
        layout = self.layout

        layout.label(text="Target:")
        layout.prop(self, "only_selected")
        layout.prop(self, "tgt_texture_size")
        layout.prop(self, "tgt_density")
        layout.prop(self, "origin")
        layout.prop(self, "tgt_area_calc_method")

        layout.separator()

    def invoke(self, context, _):
        if self.show_dialog:
            wm = context.window_manager
            return wm.invoke_props_dialog(self)

        return self.execute(context)

    def execute(self, context):
        return self.__apply_manual(context)


@BlClassRegistry()
@compat.make_annotations
class MUV_OT_WorldScaleUV_ApplyScalingDensity(bpy.types.Operator):
    """
    Operation class: Apply scaled UV (Scaling Density)
    """

    bl_idname = "uv.muv_world_scale_uv_apply_scaling_density"
    bl_label = "Apply World Scale UV (Scaling Density)"
    bl_description = "Apply scaled UV with scaling density"
    bl_options = {'REGISTER', 'UNDO'}

    tgt_scaling_factor = FloatProperty(
        name="Scaling Factor",
        default=1.0,
        max=1000.0,
        min=0.00001
    )
    origin = EnumProperty(
        name="Origin",
        description="Aspect Origin",
        items=[
            ('CENTER', "Center", "Center"),
            ('LEFT_TOP', "Left Top", "Left Bottom"),
            ('LEFT_CENTER', "Left Center", "Left Center"),
            ('LEFT_BOTTOM', "Left Bottom", "Left Bottom"),
            ('CENTER_TOP', "Center Top", "Center Top"),
            ('CENTER_BOTTOM', "Center Bottom", "Center Bottom"),
            ('RIGHT_TOP', "Right Top", "Right Top"),
            ('RIGHT_CENTER', "Right Center", "Right Center"),
            ('RIGHT_BOTTOM', "Right Bottom", "Right Bottom")

        ],
        default='CENTER'
    )
    src_density = FloatProperty(
        name="Density",
        description="Source Texel Density",
        default=0.0,
        min=0.0,
        options={'HIDDEN'}
    )
    same_density = BoolProperty(
        name="Same Density",
        description="Apply same density",
        default=False,
        options={'HIDDEN'}
    )
    show_dialog = BoolProperty(
        name="Show Diaglog Menu",
        description="Show dialog menu if true",
        default=True,
        options={'HIDDEN', 'SKIP_SAVE'}
    )
    tgt_texture = EnumProperty(
        name="Texture",
        description="Texture to be applied",
        items=_get_target_textures
    )
    tgt_area_calc_method = EnumProperty(
        name="Area Calculation Method",
        description="How to calculate target area",
        items=[
            ('MESH', "Mesh", "Calculate area by whole faces in mesh"),
            ('UV ISLAND', "UV Island", "Calculate area each UV islands"),
            ('FACE', "Face", "Calculate area each face")
        ],
        default='MESH'
    )
    only_selected = BoolProperty(
        name="Only Selected",
        description="Apply to only selected faces",
        default=True,
    )

    @classmethod
    def poll(cls, context):
        # we can not get area/space/region from console
        if common.is_console_mode():
            return True
        return _is_valid_context_for_apply(context)

    @staticmethod
    def setup_argument(ops, scene):
        ops.tgt_scaling_factor = \
            scene.muv_world_scale_uv_tgt_scaling_factor
        ops.origin = scene.muv_world_scale_uv_origin
        ops.src_density = scene.muv_world_scale_uv_src_density
        ops.same_density = False
        ops.show_dialog = False
        try:
            ops.tgt_texture = scene.muv_world_scale_uv_apply_tgt_texture
        except TypeError:
            # Workaround for the error raised when the items of EnumProperty
            # are deleted.
            ops.tgt_texture = "[Average]"
        ops.tgt_area_calc_method = \
            scene.muv_world_scale_uv_tgt_area_calc_method
        ops.only_selected = scene.muv_world_scale_uv_apply_only_selected

    def __apply_scaling_density(self, context):
        objs = common.get_uv_editable_objects(context)

        for obj in objs:
            bm = bmesh.from_edit_mesh(obj.data)
            if common.check_version(2, 73, 0) >= 0:
                bm.verts.ensure_lookup_table()
                bm.edges.ensure_lookup_table()
                bm.faces.ensure_lookup_table()

            if not bm.loops.layers.uv:
                self.report({'WARNING'},
                            "Object {} must have more than one UV map"
                            .format(obj.name))
                return {'CANCELLED'}
            uv_layer = bm.loops.layers.uv.verify()
            tex_layer = common.find_texture_layer(bm)
            faces_list = common.get_faces_list(
                bm, self.tgt_area_calc_method, self.only_selected)

            factors = []
            for faces in faces_list:
                if self.tgt_texture == "[Average]":
                    uv_area, _, density = _measure_wsuv_info_from_faces(
                        obj, bm, faces, uv_layer, tex_layer,
                        tex_selection_method='AVERAGE')
                elif self.tgt_texture == "[Max]":
                    uv_area, _, density = _measure_wsuv_info_from_faces(
                        obj, bm, faces, uv_layer, tex_layer,
                        tex_selection_method='MAX')
                elif self.tgt_texture == "[Min]":
                    uv_area, _, density = _measure_wsuv_info_from_faces(
                        obj, bm, faces, uv_layer, tex_layer,
                        tex_selection_method='MIN')
                else:
                    tgt_texture = bpy.data.images[self.tgt_texture]
                    uv_area, _, density = _measure_wsuv_info_from_faces(
                        obj, bm, faces, uv_layer, tex_layer,
                        tex_selection_method='USER_SPECIFIED',
                        tex_size=tgt_texture.size)

                if not uv_area:
                    self.report({'WARNING'},
                                "Object {} must have more than one UV map and "
                                "texture".format(obj.name))
                    return {'CANCELLED'}

                tgt_density = self.src_density * self.tgt_scaling_factor
                factor = tgt_density / density

                _apply(faces, uv_layer, self.origin, factor)
                factors.append(factor)

            bmesh.update_edit_mesh(obj.data)
            self.report({'INFO'},
                        "Scaling factor of object {}: {}"
                        .format(obj.name, factors))

        return {'FINISHED'}

    def draw(self, _):
        layout = self.layout

        layout.label(text="Source:")
        col = layout.column()
        col.prop(self, "src_density")
        col.enabled = False

        layout.separator()

        layout.label(text="Target:")
        if not self.same_density:
            layout.prop(self, "tgt_scaling_factor")
        layout.prop(self, "only_selected")
        layout.prop(self, "tgt_texture")
        layout.prop(self, "origin")
        layout.prop(self, "tgt_area_calc_method")

        layout.separator()

    def invoke(self, context, _):
        sc = context.scene

        if self.show_dialog:
            wm = context.window_manager

            if self.same_density:
                self.tgt_scaling_factor = 1.0
            else:
                self.tgt_scaling_factor = \
                    sc.muv_world_scale_uv_tgt_scaling_factor
                self.src_density = sc.muv_world_scale_uv_src_density

            return wm.invoke_props_dialog(self)

        return self.execute(context)

    def execute(self, context):
        if self.same_density:
            self.tgt_scaling_factor = 1.0

        return self.__apply_scaling_density(context)


@BlClassRegistry()
@compat.make_annotations
class MUV_OT_WorldScaleUV_ApplyProportionalToMesh(bpy.types.Operator):
    """
    Operation class: Apply scaled UV (Proportional to mesh)
    """

    bl_idname = "uv.muv_world_scale_uv_apply_proportional_to_mesh"
    bl_label = "Apply World Scale UV (Proportional to mesh)"
    bl_description = "Apply scaled UV proportionaled to mesh"
    bl_options = {'REGISTER', 'UNDO'}

    origin = EnumProperty(
        name="Origin",
        description="Aspect Origin",
        items=[
            ('CENTER', "Center", "Center"),
            ('LEFT_TOP', "Left Top", "Left Bottom"),
            ('LEFT_CENTER', "Left Center", "Left Center"),
            ('LEFT_BOTTOM', "Left Bottom", "Left Bottom"),
            ('CENTER_TOP', "Center Top", "Center Top"),
            ('CENTER_BOTTOM', "Center Bottom", "Center Bottom"),
            ('RIGHT_TOP', "Right Top", "Right Top"),
            ('RIGHT_CENTER', "Right Center", "Right Center"),
            ('RIGHT_BOTTOM', "Right Bottom", "Right Bottom")

        ],
        default='CENTER'
    )
    src_density = FloatProperty(
        name="Source Density",
        description="Source Texel Density",
        default=0.0,
        min=0.0,
        options={'HIDDEN'}
    )
    src_uv_area = FloatProperty(
        name="Source UV Area",
        description="Source UV Area",
        default=0.0,
        min=0.0,
        options={'HIDDEN'}
    )
    src_mesh_area = FloatProperty(
        name="Source Mesh Area",
        description="Source Mesh Area",
        default=0.0,
        min=0.0,
        options={'HIDDEN'}
    )
    show_dialog = BoolProperty(
        name="Show Diaglog Menu",
        description="Show dialog menu if true",
        default=True,
        options={'HIDDEN', 'SKIP_SAVE'}
    )
    tgt_texture = EnumProperty(
        name="Texture",
        description="Texture to be applied",
        items=_get_target_textures
    )
    tgt_area_calc_method = EnumProperty(
        name="Area Calculation Method",
        description="How to calculate target area",
        items=[
            ('MESH', "Mesh", "Calculate area by whole faces in mesh"),
            ('UV ISLAND', "UV Island", "Calculate area each UV islands"),
            ('FACE', "Face", "Calculate area each face")
        ],
        default='MESH'
    )
    only_selected = BoolProperty(
        name="Only Selected",
        description="Apply to only selected faces",
        default=True,
    )

    @classmethod
    def poll(cls, context):
        # we can not get area/space/region from console
        if common.is_console_mode():
            return True
        return _is_valid_context_for_apply(context)

    @staticmethod
    def setup_argument(ops, scene):
        ops.origin = scene.muv_world_scale_uv_origin
        ops.src_density = scene.muv_world_scale_uv_src_density
        ops.src_uv_area = scene.muv_world_scale_uv_src_uv_area
        ops.src_mesh_area = scene.muv_world_scale_uv_src_mesh_area
        ops.show_dialog = False
        try:
            ops.tgt_texture = scene.muv_world_scale_uv_apply_tgt_texture
        except TypeError:
            # Workaround for the error raised when the items of EnumProperty
            # are deleted.
            ops.tgt_texture = "[Average]"
        ops.tgt_area_calc_method = \
            scene.muv_world_scale_uv_tgt_area_calc_method
        ops.only_selected = scene.muv_world_scale_uv_apply_only_selected

    def __apply_proportional_to_mesh(self, context):
        objs = common.get_uv_editable_objects(context)

        for obj in objs:
            bm = bmesh.from_edit_mesh(obj.data)
            if common.check_version(2, 73, 0) >= 0:
                bm.verts.ensure_lookup_table()
                bm.edges.ensure_lookup_table()
                bm.faces.ensure_lookup_table()

            if not bm.loops.layers.uv:
                self.report({'WARNING'},
                            "Object {} must have more than one UV map"
                            .format(obj.name))
                return {'CANCELLED'}
            uv_layer = bm.loops.layers.uv.verify()
            tex_layer = common.find_texture_layer(bm)
            faces_list = common.get_faces_list(
                bm, self.tgt_area_calc_method, self.only_selected)

            factors = []
            for faces in faces_list:
                if self.tgt_texture == "[Average]":
                    uv_area, mesh_area, density = \
                        _measure_wsuv_info_from_faces(
                            obj, bm, faces, uv_layer, tex_layer,
                            tex_selection_method='AVERAGE')
                elif self.tgt_texture == "[Max]":
                    uv_area, mesh_area, density = \
                        _measure_wsuv_info_from_faces(
                            obj, bm, faces, uv_layer, tex_layer,
                            tex_selection_method='MAX')
                elif self.tgt_texture == "[Min]":
                    uv_area, mesh_area, density = \
                        _measure_wsuv_info_from_faces(
                            obj, bm, faces, uv_layer, tex_layer,
                            tex_selection_method='MIN')
                else:
                    tgt_texture = bpy.data.images[self.tgt_texture]
                    uv_area, mesh_area, density = \
                        _measure_wsuv_info_from_faces(
                            obj, bm, faces, uv_layer, tex_layer,
                            tex_selection_method='USER_SPECIFIED',
                            tex_size=tgt_texture.size)
                if not uv_area:
                    self.report({'WARNING'},
                                "Object {} must have more than one UV map and "
                                "texture".format(obj.name))
                    return {'CANCELLED'}

                tgt_density = self.src_density * sqrt(mesh_area) / sqrt(
                    self.src_mesh_area)
                factor = tgt_density / density

                _apply(faces, uv_layer, self.origin, factor)
                factors.append(factor)

            bmesh.update_edit_mesh(obj.data)
            self.report({'INFO'},
                        "Scaling factor of object {}: {}"
                        .format(obj.name, factors))

        return {'FINISHED'}

    def draw(self, _):
        layout = self.layout

        layout.label(text="Source:")
        col = layout.column(align=True)
        col.prop(self, "src_density")
        col.prop(self, "src_uv_area")
        col.prop(self, "src_mesh_area")
        col.enabled = False

        layout.separator()

        layout.label(text="Target:")
        layout.prop(self, "only_selected")
        layout.prop(self, "origin")
        layout.prop(self, "tgt_area_calc_method")
        layout.prop(self, "tgt_texture")

        layout.separator()

    def invoke(self, context, _):
        if self.show_dialog:
            wm = context.window_manager
            sc = context.scene

            self.src_density = sc.muv_world_scale_uv_src_density
            self.src_mesh_area = sc.muv_world_scale_uv_src_mesh_area

            return wm.invoke_props_dialog(self)

        return self.execute(context)

    def execute(self, context):
        return self.__apply_proportional_to_mesh(context)



# Copyright 2018-2021 The glTF-Blender-IO authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import bpy
import os
import tempfile
from os.path import dirname, join, isfile, basename, normpath
import urllib.parse
import re

from ...io.imp.gltf2_io_binary import BinaryData


# Note that Image is not a glTF2.0 object
class BlenderImage():
    """Manage Image."""
    def __new__(cls, *args, **kwargs):
        raise RuntimeError("%s should not be instantiated" % cls)

    @staticmethod
    def create(gltf, img_idx):
        """Image creation."""
        img = gltf.data.images[img_idx]
        img_name = img.name

        if img.blender_image_name is not None:
            # Image is already used somewhere
            return

        tmp_dir = None
        is_placeholder = False
        try:
            if img.uri is not None and not img.uri.startswith('data:'):
                # Image stored in a file
                path = join(dirname(gltf.filename), _uri_to_path(img.uri))
                img_name = img_name or basename(path)

            else:
                # Image stored as data => create a tempfile, pack, and delete file
                img_data = BinaryData.get_image_data(gltf, img_idx)
                if img_data is None:
                    return
                img_name = img_name or 'Image_%d' % img_idx
                tmp_dir = tempfile.TemporaryDirectory(prefix='gltfimg-')
                filename = _filenamify(img_name) or 'Image_%d' % img_idx
                filename += _img_extension(img)
                path = join(tmp_dir.name, filename)
                with open(path, 'wb') as f:
                    f.write(img_data)

            num_images = len(bpy.data.images)

            try:
                blender_image = bpy.data.images.load(
                    os.path.abspath(path),
                    check_existing=tmp_dir is None,
                )
            except RuntimeError:
                gltf.log.error("Missing image file (index %d): %s" % (img_idx, path))
                blender_image = _placeholder_image(img_name, os.path.abspath(path))
                is_placeholder = True

            if len(bpy.data.images) != num_images:  # If created a new image
                blender_image.name = img_name

                needs_pack = (
                    gltf.import_settings['import_pack_images'] or
                    tmp_dir is not None
                )
                if not is_placeholder and needs_pack:
                    blender_image.pack()

            img.blender_image_name = blender_image.name

        finally:
            if tmp_dir is not None:
                tmp_dir.cleanup()

def _placeholder_image(name, path):
    image = bpy.data.images.new(name, 128, 128)
    # allow the path to be resolved later
    image.filepath = path
    image.source = 'FILE'
    return image

def _uri_to_path(uri):
    uri = urllib.parse.unquote(uri)
    return normpath(uri)

def _img_extension(img):
    if img.mime_type == 'image/png':
        return '.png'
    if img.mime_type == 'image/jpeg':
        return '.jpg'
    return ''

def _filenamify(s):
    s = s.strip().replace(' ', '_')
    return re.sub(r'(?u)[^-\w.]', '', s)



# <pep8-80 compliant>

# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

__author__ = "Nutti <nutti.metro@gmail.com>"
__status__ = "production"
__version__ = "6.5"
__date__ = "6 Mar 2021"

import bpy
from bpy.props import BoolProperty, FloatProperty, EnumProperty
import bmesh

from .. import common
from ..utils.bl_class_registry import BlClassRegistry
from ..utils.property_class_registry import PropertyClassRegistry
from ..utils import compatibility as compat


def _is_valid_context(context):
    objs = common.get_uv_editable_objects(context)
    if not objs:
        return False

    # only edit mode is allowed to execute
    if context.object.mode != 'EDIT':
        return False

    # 'IMAGE_EDITOR' and 'VIEW_3D' space is allowed to execute.
    # If 'View_3D' space is not allowed, you can't find option in Tool-Shelf
    # after the execution
    if not common.is_valid_space(context, ['IMAGE_EDITOR', 'VIEW_3D']):
        return False

    return True


@PropertyClassRegistry()
class _Properties:
    idname = "select_uv"

    @classmethod
    def init_props(cls, scene):
        scene.muv_select_uv_enabled = BoolProperty(
            name="Select UV Enabled",
            description="Select UV is enabled",
            default=False
        )
        scene.muv_select_uv_same_polygon_threshold = FloatProperty(
            name="Same Polygon Threshold",
            description="Threshold to distinguish same polygons",
            default=0.000001,
            min=0.000001,
            max=0.01,
            step=0.00001
        )
        scene.muv_select_uv_selection_method = EnumProperty(
            name="Selection Method",
            description="How to select faces which have overlapped UVs",
            items=[
                ('EXTEND', "Extend",
                 "Select faces without unselecting selected faces"),
                ('RESET', "Reset", "Select faces and unselect selected faces"),
            ],
            default='RESET'
        )
        scene.muv_select_uv_sync_mesh_selection = BoolProperty(
            name="Sync Mesh Selection",
            description="Select the mesh's faces as well as UV's faces",
            default=False
        )

    @classmethod
    def del_props(cls, scene):
        del scene.muv_select_uv_enabled
        del scene.muv_select_uv_same_polygon_threshold


@BlClassRegistry()
@compat.make_annotations
class MUV_OT_SelectUV_SelectOverlapped(bpy.types.Operator):
    """
    Operation class: Select faces which have overlapped UVs
    """

    bl_idname = "uv.muv_select_uv_select_overlapped"
    bl_label = "Overlapped"
    bl_description = "Select faces which have overlapped UVs"
    bl_options = {'REGISTER', 'UNDO'}

    same_polygon_threshold = FloatProperty(
        name="Same Polygon Threshold",
        description="Threshold to distinguish same polygons",
        default=0.000001,
        min=0.000001,
        max=0.01,
        step=0.00001
    )
    selection_method = EnumProperty(
        name="Selection Method",
        description="How to select faces which have overlapped UVs",
        items=[
            ('EXTEND', "Extend",
             "Select faces without unselecting selected faces"),
            ('RESET', "Reset", "Select faces and unselect selected faces"),
        ],
        default='RESET'
    )
    sync_mesh_selection = BoolProperty(
        name="Sync Mesh Selection",
        description="Select mesh's faces as well as UV's faces",
        default=False
    )

    @classmethod
    def poll(cls, context):
        # we can not get area/space/region from console
        if common.is_console_mode():
            return True
        return _is_valid_context(context)

    @staticmethod
    def setup_argument(ops, scene):
        ops.same_polygon_threshold = scene.muv_select_uv_same_polygon_threshold
        ops.selection_method = scene.muv_select_uv_selection_method
        ops.sync_mesh_selection = scene.muv_select_uv_sync_mesh_selection

    def execute(self, context):
        objs = common.get_uv_editable_objects(context)

        bm_list = []
        uv_layer_list = []
        faces_list = []
        for obj in objs:
            bm = bmesh.from_edit_mesh(obj.data)
            if common.check_version(2, 73, 0) >= 0:
                bm.faces.ensure_lookup_table()
            uv_layer = bm.loops.layers.uv.verify()

            if context.tool_settings.use_uv_select_sync:
                sel_faces = [f for f in bm.faces]
            else:
                sel_faces = [f for f in bm.faces if f.select]
            bm_list.append(bm)
            uv_layer_list.append(uv_layer)
            faces_list.append(sel_faces)

        overlapped_info = common.get_overlapped_uv_info(
            bm_list, faces_list, uv_layer_list, 'FACE',
            self.same_polygon_threshold)

        if self.selection_method == 'RESET':
            if context.tool_settings.use_uv_select_sync:
                for faces in faces_list:
                    for f in faces:
                        f.select = False
            else:
                for uv_layer, faces in zip(uv_layer_list, faces_list):
                    for f in faces:
                        if self.sync_mesh_selection:
                            f.select = False
                        for l in f.loops:
                            l[uv_layer].select = False

        for info in overlapped_info:
            if context.tool_settings.use_uv_select_sync:
                info["subject_face"].select = True
            else:
                if self.sync_mesh_selection:
                    info["subject_face"].select = True
                for l in info["subject_face"].loops:
                    l[info["subject_uv_layer"]].select = True

        for obj in objs:
            bmesh.update_edit_mesh(obj.data)

        return {'FINISHED'}


@BlClassRegistry()
@compat.make_annotations
class MUV_OT_SelectUV_SelectFlipped(bpy.types.Operator):
    """
    Operation class: Select faces which have flipped UVs
    """

    bl_idname = "uv.muv_select_uv_select_flipped"
    bl_label = "Flipped"
    bl_description = "Select faces which have flipped UVs"
    bl_options = {'REGISTER', 'UNDO'}

    selection_method = EnumProperty(
        name="Selection Method",
        description="How to select faces which have overlapped UVs",
        items=[
            ('EXTEND', "Extend",
             "Select faces without unselecting selected faces"),
            ('RESET', "Reset", "Select faces and unselect selected faces"),
        ],
        default='RESET'
    )
    sync_mesh_selection = BoolProperty(
        name="Sync Mesh Selection",
        description="Select mesh's faces as well as UV's faces",
        default=False
    )

    @classmethod
    def poll(cls, context):
        # we can not get area/space/region from console
        if common.is_console_mode():
            return True
        return _is_valid_context(context)

    @staticmethod
    def setup_argument(ops, scene):
        ops.selection_method = scene.muv_select_uv_selection_method
        ops.sync_mesh_selection = scene.muv_select_uv_sync_mesh_selection

    def execute(self, context):
        objs = common.get_uv_editable_objects(context)

        bm_list = []
        uv_layer_list = []
        faces_list = []
        for obj in objs:
            bm = bmesh.from_edit_mesh(obj.data)
            if common.check_version(2, 73, 0) >= 0:
                bm.faces.ensure_lookup_table()
            uv_layer = bm.loops.layers.uv.verify()

            if context.tool_settings.use_uv_select_sync:
                sel_faces = [f for f in bm.faces]
            else:
                sel_faces = [f for f in bm.faces if f.select]
            bm_list.append(bm)
            uv_layer_list.append(uv_layer)
            faces_list.append(sel_faces)

        flipped_info = common.get_flipped_uv_info(
            bm_list, faces_list, uv_layer_list)

        if self.selection_method == 'RESET':
            if context.tool_settings.use_uv_select_sync:
                for faces in faces_list:
                    for f in faces:
                        f.select = False
            else:
                for uv_layer, faces in zip(uv_layer_list, faces_list):
                    for f in faces:
                        if self.sync_mesh_selection:
                            f.select = False
                        for l in f.loops:
                            l[uv_layer].select = False

        for info in flipped_info:
            if context.tool_settings.use_uv_select_sync:
                info["face"].select = True
            else:
                if self.sync_mesh_selection:
                    info["face"].select = True
                for l in info["face"].loops:
                    l[info["uv_layer"]].select = True

        for obj in objs:
            bmesh.update_edit_mesh(obj.data)

        return {'FINISHED'}


@BlClassRegistry()
class MUV_OT_SelectUV_ZoomSelectedUV(bpy.types.Operator):
    """
    Operation class: Zoom selected UV in View3D space
    """

    bl_idname = "uv.muv_select_uv_zoom_selected_uv"
    bl_label = "Zoom Selected UV"
    bl_description = "Zoom selected UV in View3D space"
    bl_options = {'REGISTER', 'UNDO'}

    @classmethod
    def poll(cls, context):
        # we can not get area/space/region from console
        if common.is_console_mode():
            return True
        return _is_valid_context(context)

    def _get_override_context(self, context):
        for window in context.window_manager.windows:
            screen = window.screen
            for area in screen.areas:
                if area.type == 'VIEW_3D':
                    for region in area.regions:
                        if region.type == 'WINDOW':
                            return {'window': window, 'screen': screen,
                                    'area': area, 'region': region}
        return None

    def execute(self, context):
        objs = common.get_uv_editable_objects(context)

        bm_list = []
        uv_layer_list = []
        verts_list = []
        for obj in objs:
            bm = bmesh.from_edit_mesh(obj.data)
            if common.check_version(2, 73, 0) >= 0:
                bm.verts.ensure_lookup_table()
            uv_layer = bm.loops.layers.uv.verify()

            sel_verts = [v for v in bm.verts if v.select]
            bm_list.append(bm)
            uv_layer_list.append(uv_layer)
            verts_list.append(sel_verts)

        # Get all selected UV vertices in UV Editor.
        sel_uv_verts = []
        for vlist, uv_layer in zip(verts_list, uv_layer_list):
            for v in vlist:
                for l in v.link_loops:
                    if l[uv_layer].select or \
                            context.tool_settings.use_uv_select_sync:
                        sel_uv_verts.append(v)
                        break

        # Select vertices only selected in UV Editor.
        for bm in bm_list:
            for v in bm.verts:
                v.select = False
        for v in sel_uv_verts:
            v.select = True
        for obj in objs:
            bmesh.update_edit_mesh(obj.data)

        # Zoom.
        override_context = self._get_override_context(context)
        if override_context is None:
            self.report({'WARNING'}, "More than one 'VIEW_3D' area must exist")
            return {'CANCELLED'}
        bpy.ops.view3d.view_selected(override_context, use_all_regions=False)

        # Revert selection of verticies.
        for v in sel_verts:
            v.select = True
        for obj in objs:
            bmesh.update_edit_mesh(obj.data)

        return {'FINISHED'}



#
# Copyright 2011-2013 Blender Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# <pep8 compliant>
from __future__ import annotations


def _is_using_buggy_driver():
    import bgl
    # We need to be conservative here because in multi-GPU systems display card
    # might be quite old, but others one might be just good.
    #
    # So We shouldn't disable possible good dedicated cards just because display
    # card seems weak. And instead we only blacklist configurations which are
    # proven to cause problems.
    if bgl.glGetString(bgl.GL_VENDOR) == "ATI Technologies Inc.":
        import re
        version = bgl.glGetString(bgl.GL_VERSION)
        if version.endswith("Compatibility Profile Context"):
            # Old HD 4xxx and 5xxx series drivers did not have driver version
            # in the version string, but those cards do not quite work and
            # causing crashes.
            return True
        regex = re.compile(".*Compatibility Profile Context ([0-9]+(\\.[0-9]+)+)$")
        if not regex.match(version):
            # Skip cards like FireGL
            return False
        version = regex.sub("\\1", version).split('.')
        return int(version[0]) == 8
    return False


def _workaround_buggy_drivers():
    if _is_using_buggy_driver():
        import _cycles
        if hasattr(_cycles, "opencl_disable"):
            print("Cycles: OpenGL driver known to be buggy, disabling OpenCL platform.")
            _cycles.opencl_disable()


def _configure_argument_parser():
    import argparse
    # No help because it conflicts with general Python scripts argument parsing
    parser = argparse.ArgumentParser(description="Cycles Addon argument parser",
                                     add_help=False)
    parser.add_argument("--cycles-resumable-num-chunks",
                        help="Number of chunks to split sample range into",
                        default=None)
    parser.add_argument("--cycles-resumable-current-chunk",
                        help="Current chunk of samples range to render",
                        default=None)
    parser.add_argument("--cycles-resumable-start-chunk",
                        help="Start chunk to render",
                        default=None)
    parser.add_argument("--cycles-resumable-end-chunk",
                        help="End chunk to render",
                        default=None)
    parser.add_argument("--cycles-print-stats",
                        help="Print rendering statistics to stderr",
                        action='store_true')
    parser.add_argument("--cycles-device",
                        help="Set the device to use for Cycles, overriding user preferences and the scene setting."
                             "Valid options are 'CPU', 'CUDA', 'OPTIX' or 'OPENCL'."
                             "Additionally, you can append '+CPU' to any GPU type for hybrid rendering.",
                        default=None)
    return parser


def _parse_command_line():
    import sys

    argv = sys.argv
    if "--" not in argv:
        return

    parser = _configure_argument_parser()
    args, _ = parser.parse_known_args(argv[argv.index("--") + 1:])

    if args.cycles_resumable_num_chunks is not None:
        if args.cycles_resumable_current_chunk is not None:
            import _cycles
            _cycles.set_resumable_chunk(
                int(args.cycles_resumable_num_chunks),
                int(args.cycles_resumable_current_chunk),
            )
        elif args.cycles_resumable_start_chunk is not None and \
                args.cycles_resumable_end_chunk:
            import _cycles
            _cycles.set_resumable_chunk_range(
                int(args.cycles_resumable_num_chunks),
                int(args.cycles_resumable_start_chunk),
                int(args.cycles_resumable_end_chunk),
            )
    if args.cycles_print_stats:
        import _cycles
        _cycles.enable_print_stats()

    if args.cycles_device:
        import _cycles
        _cycles.set_device_override(args.cycles_device)


def init():
    import bpy
    import _cycles
    import os.path

    # Workaround possibly buggy legacy drivers which crashes on the OpenCL
    # device enumeration.
    #
    # This checks are not really correct because they might still fail
    # in the case of multiple GPUs. However, currently buggy drivers
    # are really old and likely to be used in single GPU systems only
    # anyway.
    #
    # Can't do it in the background mode, so we hope OpenCL is no enabled
    # in the user preferences.
    if not bpy.app.background:
        _workaround_buggy_drivers()

    path = os.path.dirname(__file__)
    user_path = os.path.dirname(os.path.abspath(bpy.utils.user_resource('CONFIG', '')))

    _cycles.init(path, user_path, bpy.app.background)
    _parse_command_line()


def exit():
    import _cycles
    _cycles.exit()


def create(engine, data, region=None, v3d=None, rv3d=None, preview_osl=False):
    import _cycles
    import bpy

    data = data.as_pointer()
    prefs = bpy.context.preferences.as_pointer()
    screen = 0
    if region:
        screen = region.id_data.as_pointer()
        region = region.as_pointer()
    if v3d:
        screen = screen or v3d.id_data.as_pointer()
        v3d = v3d.as_pointer()
    if rv3d:
        screen = screen or rv3d.id_data.as_pointer()
        rv3d = rv3d.as_pointer()

    engine.session = _cycles.create(engine.as_pointer(), prefs, data, screen, region, v3d, rv3d, preview_osl)


def free(engine):
    if hasattr(engine, "session"):
        if engine.session:
            import _cycles
            _cycles.free(engine.session)
        del engine.session


def render(engine, depsgraph):
    import _cycles
    if hasattr(engine, "session"):
        _cycles.render(engine.session, depsgraph.as_pointer())


def bake(engine, depsgraph, obj, pass_type, pass_filter, width, height):
    import _cycles
    session = getattr(engine, "session", None)
    if session is not None:
        _cycles.bake(engine.session, depsgraph.as_pointer(), obj.as_pointer(), pass_type, pass_filter, width, height)


def reset(engine, data, depsgraph):
    import _cycles
    import bpy

    prefs = bpy.context.preferences
    if prefs.experimental.use_cycles_debug and prefs.view.show_developer_ui:
        _cycles.debug_flags_update(depsgraph.scene.as_pointer())
    else:
        _cycles.debug_flags_reset()

    data = data.as_pointer()
    depsgraph = depsgraph.as_pointer()
    _cycles.reset(engine.session, data, depsgraph)


def sync(engine, depsgraph, data):
    import _cycles
    _cycles.sync(engine.session, depsgraph.as_pointer())


def draw(engine, depsgraph, region, v3d, rv3d):
    import _cycles
    depsgraph = depsgraph.as_pointer()
    v3d = v3d.as_pointer()
    rv3d = rv3d.as_pointer()

    # draw render image
    _cycles.draw(engine.session, depsgraph, v3d, rv3d)


def available_devices():
    import _cycles
    return _cycles.available_devices()


def with_osl():
    import _cycles
    return _cycles.with_osl


def with_network():
    import _cycles
    return _cycles.with_network


def system_info():
    import _cycles
    return _cycles.system_info()


def list_render_passes(scene, srl):
    # Builtin Blender passes.
    yield ("Combined", "RGBA", 'COLOR')

    if srl.use_pass_z:                     yield ("Depth",         "Z",    'VALUE')
    if srl.use_pass_mist:                  yield ("Mist",          "Z",    'VALUE')
    if srl.use_pass_normal:                yield ("Normal",        "XYZ",  'VECTOR')
    if srl.use_pass_vector:                yield ("Vector",        "XYZW", 'VECTOR')
    if srl.use_pass_uv:                    yield ("UV",            "UVA",  'VECTOR')
    if srl.use_pass_object_index:          yield ("IndexOB",       "X",    'VALUE')
    if srl.use_pass_material_index:        yield ("IndexMA",       "X",    'VALUE')
    if srl.use_pass_shadow:                yield ("Shadow",        "RGB",  'COLOR')
    if srl.use_pass_ambient_occlusion:     yield ("AO",            "RGB",  'COLOR')
    if srl.use_pass_diffuse_direct:        yield ("DiffDir",       "RGB",  'COLOR')
    if srl.use_pass_diffuse_indirect:      yield ("DiffInd",       "RGB",  'COLOR')
    if srl.use_pass_diffuse_color:         yield ("DiffCol",       "RGB",  'COLOR')
    if srl.use_pass_glossy_direct:         yield ("GlossDir",      "RGB",  'COLOR')
    if srl.use_pass_glossy_indirect:       yield ("GlossInd",      "RGB",  'COLOR')
    if srl.use_pass_glossy_color:          yield ("GlossCol",      "RGB",  'COLOR')
    if srl.use_pass_transmission_direct:   yield ("TransDir",      "RGB",  'COLOR')
    if srl.use_pass_transmission_indirect: yield ("TransInd",      "RGB",  'COLOR')
    if srl.use_pass_transmission_color:    yield ("TransCol",      "RGB",  'COLOR')
    if srl.use_pass_emit:                  yield ("Emit",          "RGB",  'COLOR')
    if srl.use_pass_environment:           yield ("Env",           "RGB",  'COLOR')

    # Cycles specific passes.
    crl = srl.cycles
    if crl.pass_debug_render_time:             yield ("Debug Render Time",             "X",   'VALUE')
    if crl.pass_debug_bvh_traversed_nodes:     yield ("Debug BVH Traversed Nodes",     "X",   'VALUE')
    if crl.pass_debug_bvh_traversed_instances: yield ("Debug BVH Traversed Instances", "X",   'VALUE')
    if crl.pass_debug_bvh_intersections:       yield ("Debug BVH Intersections",       "X",   'VALUE')
    if crl.pass_debug_ray_bounces:             yield ("Debug Ray Bounces",             "X",   'VALUE')
    if crl.pass_debug_sample_count:            yield ("Debug Sample Count",            "X",   'VALUE')
    if crl.use_pass_volume_direct:             yield ("VolumeDir",                     "RGB", 'COLOR')
    if crl.use_pass_volume_indirect:           yield ("VolumeInd",                     "RGB", 'COLOR')

    # Cryptomatte passes.
    crypto_depth = (srl.pass_cryptomatte_depth + 1) // 2
    if srl.use_pass_cryptomatte_object:
        for i in range(0, crypto_depth):
            yield ("CryptoObject" + '{:02d}'.format(i), "RGBA", 'COLOR')
    if srl.use_pass_cryptomatte_material:
        for i in range(0, crypto_depth):
            yield ("CryptoMaterial" + '{:02d}'.format(i), "RGBA", 'COLOR')
    if srl.use_pass_cryptomatte_asset:
        for i in range(0, crypto_depth):
            yield ("CryptoAsset" + '{:02d}'.format(i), "RGBA", 'COLOR')

    # Denoising passes.
    if (scene.cycles.use_denoising and crl.use_denoising) or crl.denoising_store_passes:
        yield ("Noisy Image", "RGBA", 'COLOR')
        if crl.denoising_store_passes:
            yield ("Denoising Normal",          "XYZ", 'VECTOR')
            yield ("Denoising Albedo",          "RGB", 'COLOR')
            yield ("Denoising Depth",           "Z",   'VALUE')

            if scene.cycles.denoiser == 'NLM':
                yield ("Denoising Shadowing",       "X",   'VALUE')
                yield ("Denoising Variance",        "RGB", 'COLOR')
                yield ("Denoising Intensity",       "X",   'VALUE')

                clean_options = ("denoising_diffuse_direct", "denoising_diffuse_indirect",
                                 "denoising_glossy_direct", "denoising_glossy_indirect",
                                 "denoising_transmission_direct", "denoising_transmission_indirect")
                if any(getattr(crl, option) for option in clean_options):
                    yield ("Denoising Clean", "RGB", 'COLOR')

    # Custom AOV passes.
    for aov in srl.aovs:
        if aov.type == 'VALUE':
            yield (aov.name, "X", 'VALUE')
        else:
            yield (aov.name, "RGBA", 'COLOR')


def register_passes(engine, scene, view_layer):
    for name, channelids, channeltype in list_render_passes(scene, view_layer):
        engine.register_pass(scene, view_layer, name, len(channelids), channelids, channeltype)
